# -*- coding: utf-8 -*-

# Copyright (C) 2006 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous functions."""

import aeidon
import collections
import contextlib
import inspect
import locale
import mimetypes
import os
import random
import re
import shutil
import stat
import subprocess
import sys
import traceback
import urllib.parse

VIDEO_FILE_EXTENSIONS = [
    ".avi",
    ".flv",
    ".mkv",
    ".mov",
    ".mp4",
    ".ogg",
    ".ogv",
    ".vob",
    ".webm",
]


def affirm(value):
    """Raise :exc:`aeidon.AffirmationError` if value evaluates to ``False``."""
    if not value:
        raise aeidon.AffirmationError("Not True: {!r}".format(value))

@contextlib.contextmanager
def atomic_open(path, mode="w", *args, **kwargs):
    """
    A context manager for atomically writing a file.

    The file is written to a temporary file on the same filesystem, flushed and
    fsynced and then renamed to replace the existing file. This should
    (probably) be atomic on any Unix system. On Windows, it should (probably)
    be atomic if using Python 3.3 or greater.
    """
    path = os.path.realpath(path)
    chars = list("abcdefghijklmnopqrstuvwxyz0123456789")
    directory = os.path.dirname(path)
    basename = os.path.basename(path)
    while True:
        # Let's use a hidden temporary file to avoid a file
        # flickering in a possibly open file browser window.
        suffix = "".join(random.sample(chars, 8))
        temp_basename = ".{}.tmp{}".format(basename, suffix)
        temp_path = os.path.join(directory, temp_basename)
        if not os.path.isfile(temp_path): break
    try:
        if os.path.isfile(path):
            # If the file exists, use the same permissions.
            # Note that all other file metadata, including
            # owner and group, is not preserved.
            with open(temp_path, "w") as f: pass
            st = os.stat(path)
            os.chmod(temp_path, stat.S_IMODE(st.st_mode))
        with open(temp_path, mode, *args, **kwargs) as f:
            yield f
            f.flush()
            os.fsync(f.fileno())
        try:
            if hasattr(os, "replace"):
                # os.replace was added in Python 3.3.
                # This should be atomic on Windows too.
                os.replace(temp_path, path)
            else:
                # os.rename is atomic on Unix, but fails
                # on Windows if the file exists.
                os.rename(temp_path, path)
            # os.rename and os.replace will fail if path
            # and temp_path are not on the same device,
            # for instance they can be on two separate
            # branches of a union mount. Atomicity is not
            # possible in this case.
        except OSError:
            # Fall back to a non-atomic operation using
            # shutil.move. On Windows this requires that
            # the destination file does not exist.
            if sys.platform == "win32":
                if os.path.isfile(path):
                    os.remove(path)
            shutil.move(temp_path, path)
    finally:
        with silent(Exception):
            os.remove(temp_path)

@aeidon.deco.once
def chardet_available():
    """Return ``True`` if :mod:`chardet` module is available."""
    try:
        import chardet # noqa
        return True
    except Exception:
        return False

def compare_versions(x, y):
    """
    Compare version strings `x` and `y`.

    Used version number formats are ``MAJOR.MINOR``, ``MAJOR.MINOR.PATCH``
    and ``MAJOR.MINOR.PATCH.DATE/REVISION``, where all items are integers.
    Return 1 if `x` newer, 0 if equal or -1 if `y` newer.
    """
    compare = lambda a, b: (a > b) - (a < b)
    return compare(list(map(int, x.split("."))),
                   list(map(int, y.split("."))))

def connect(observer, observable, signal, *args):
    """
    Connect `observable`'s `signal` to `observer`'s callback method.

    If `observable` is a string, it should be an attribute of `observer`.
    If `observable` is not a string it should be the same as `observer`.
    """
    method_name = signal.replace("-", "_").replace("::", "_")
    if observer is not observable:
        method_name = "_".join((observable, method_name))
    method_name = "_on_{}".format(method_name).replace("__", "_")
    if not hasattr(observer, method_name):
        method_name = method_name[1:]
    method = getattr(observer, method_name)
    if observer is not observable:
        observable = getattr(observer, observable)
    return observable.connect(signal, method, *args)

def detect_format(path, encoding):
    """
    Detect and return format of subtitle file at `path`.

    Raise :exc:`IOError` if reading fails.
    Raise :exc:`UnicodeError` if decoding fails.
    Raise :exc:`aeidon.FormatError` if unable to detect format.
    Return an :attr:`aeidon.formats` enumeration item.
    """
    re_ids = [(x, re.compile(x.identifier)) for x in aeidon.formats]
    with open(path, "r", encoding=encoding) as f:
        for line in f:
            for format, re_id in re_ids:
                if re_id.search(line) is not None:
                    return format
    raise aeidon.FormatError("Failed to detect format of file {!r}"
                             .format(path))

def detect_newlines(path):
    """Detect and return the newline type of file at `path` or ``None``."""
    try:
        with open(path, "r", newline="") as f:
            f.read()
            chars = f.newlines
    except Exception:
        return None
    if chars is None:
        return None
    if isinstance(chars, str):
        return aeidon.newlines.find_item("value", chars)
    if isinstance(chars, tuple):
        if len(chars) == 1:
            return aeidon.newlines.find_item("value", chars[0])
        # This is not actually correct. If both CR and LF are detected,
        # it could mean a mixture of Mac and Unix newlines on separate
        # lines or one Windows newline in a mostly something else file.
        # We could count the frequencies, but it's probably not worth
        # the effort.
        return aeidon.newlines.WINDOWS
    return None

def flatten(lst):
    """
    Return a shallow version of `lst`.

    >>> aeidon.util.flatten([1, 2, 3, [4, 5, [6]]])
    [1, 2, 3, 4, 5, 6]
    """
    flat_lst = []
    for item in lst:
        if isinstance(item, list):
            flat_lst.extend(flatten(item))
        else: # Non-list item.
            flat_lst.append(item)
    return flat_lst

def get_chardet_version():
    """Return :mod:`chardet` version number as string or ``None``."""
    try:
        import chardet
        return chardet.__version__
    except Exception:
        return None

@aeidon.deco.once
def get_default_encoding(fallback="utf_8"):
    """Return the locale encoding or `fallback`."""
    encoding = locale.getpreferredencoding()
    encoding = encoding or fallback
    re_illegal = re.compile(r"[^a-z0-9_]")
    encoding = re_illegal.sub("_", encoding.lower())
    encoding = get_encoding_alias(encoding)
    return encoding

@aeidon.deco.once
def get_default_newline():
    """Return system default newline as :attr:`aeidon.newlines` item."""
    return aeidon.newlines.find_item("value", os.linesep)

def get_encoding_alias(encoding):
    """Return proper Python alias for `encoding`."""
    from encodings.aliases import aliases
    with silent(LookupError):
        return aliases[encoding]
    return encoding

def get_ranges(lst):
    """
    Return a list of ranges in list of integers.

    >>> aeidon.util.get_ranges([1, 2, 3, 5, 6, 7, 9, 11, 12])
    [[1, 2, 3], [5, 6, 7], [9], [11, 12]]
    """
    if not lst: return []
    lst = sorted(get_unique(lst))
    ranges = [[lst.pop(0)]]
    for item in lst:
        if item == ranges[-1][-1] + 1:
            ranges[-1].append(item)
        else:
            ranges.append([item])
    return ranges

def get_template_header(format):
    """
    Read and return the template header for `format`.

    Raise :exc:`IOError` if reading global header file fails.
    Raise :exc:`UnicodeError` if decoding global header file fails.
    """
    directory = os.path.join(aeidon.DATA_HOME_DIR, "headers")
    path = os.path.join(directory, format.name.lower())
    with silent(Exception):
        header = read(path, encoding=None, quiet=True).rstrip()
        return normalize_newlines(header)
    directory = os.path.join(aeidon.DATA_DIR, "headers")
    path = os.path.join(directory, format.name.lower())
    header = read(path, "ascii").rstrip()
    return normalize_newlines(header)

def get_unique(lst, keep_last=False):
    """Return `lst` with duplicates removed."""
    if keep_last:
        return list(reversed(get_unique(list(reversed(lst)))))
    # https://stackoverflow.com/a/7961425
    return list(collections.OrderedDict.fromkeys(lst))

def install_module(name, obj):
    """
    Install `obj`'s module into the :mod:`aeidon` namespace.

    Typical call is of form::

        aeidon.util.install_module("foo", lambda: None)
    """
    aeidon.__dict__[name] = inspect.getmodule(obj)

def is_video_file(path):
    """Return ``True`` if `path` is a video file."""
    if not os.path.isfile(path):
        return False
    # The mimetypes module doesn't work well on Windows,
    # fall back on a custom list of video file extensions.
    type, encoding = mimetypes.guess_type(path)
    return ((type and type.startswith("video/")) or
            path.lower().endswith(tuple(VIDEO_FILE_EXTENSIONS)))

def last(iterator):
    """Return the last value from `iterator` or ``None``."""
    value = None
    for value in iterator: pass
    return value

def makedirs(directory):
    """
    Recursively make `directory` if it does not exist.

    Raise :exc:`OSError` if unsuccessful.
    """
    directory = os.path.abspath(directory)
    if os.path.isdir(directory):
        return directory
    try:
        os.makedirs(directory)
    except OSError as error:
        print("Failed to create directory {!r}: {!s}"
              .format(directory, error),
              file=sys.stderr)
        raise # OSError
    return directory

def normalize_newlines(text):
    """Convert all newlines in `text` to "\\n"."""
    re_newline_char = re.compile(r"\r\n?")
    return re_newline_char.sub("\n", text)

def path_to_uri(path):
    """Convert local filepath to URI."""
    if sys.platform == "win32":
        path = "/{}".format(path.replace("\\", "/"))
    return "file://{}".format(urllib.parse.quote(path))

def print_read_io(exc_info, path):
    """Print :exc:`IOError` message to standard error."""
    print("Failed to read file '{}': {}"
          .format(path, exc_info[1].args[1]),
          file=sys.stderr)

def print_read_unicode(exc_info, path, encoding):
    """Print :exc:`UnicodeError` message to standard error."""
    encoding = encoding or get_default_encoding()
    print("Failed to decode file '{}' with codec '{}'"
          .format(path, encoding),
          file=sys.stderr)

def print_write_io(exc_info, path):
    """Print :exc:`IOError` message to standard error."""
    print("Failed to write file '{}': {}"
          .format(path, exc_info[1].args[1]),
          file=sys.stderr)

def print_write_unicode(exc_info, path, encoding):
    """Print :exc:`UnicodeError` message to standard error."""
    encoding = encoding or get_default_encoding()
    print("Failed to encode file '{}' with codec '{}'"
          .format(path, encoding),
          file=sys.stderr)

def read(path, encoding=None, fallback="utf_8", quiet=False):
    """
    Read file at `path` and return text.

    `fallback` should be ``None`` to try only `encoding`.
    Raise :exc:`IOError` if reading fails.
    Raise :exc:`UnicodeError` if decoding fails.
    """
    encoding = encoding or get_default_encoding()
    try:
        with open(path, "r", encoding=encoding) as f:
            return f.read().strip()
    except IOError:
        if not quiet:
            print_read_io(sys.exc_info(), path)
        raise # IOError
    except UnicodeError:
        if not fallback in (encoding, None, ""):
            return read(path, fallback, None, quiet)
        if not quiet:
            print_read_unicode(sys.exc_info(), path, encoding)
        raise # UnicodeError

def readlines(path, encoding=None, fallback="utf_8", quiet=False):
    """
    Read file at `path` and return lines.

    `fallback` should be ``None`` to try only `encoding`.
    Raise :exc:`IOError` if reading fails.
    Raise :exc:`UnicodeError` if decoding fails.
    """
    text = read(path, encoding, fallback, quiet)
    text = normalize_newlines(text)
    return text.split("\n")

def replace_extension(path, format):
    """Replace possible extension in `path` with that of `format`."""
    extensions = [x.extension for x in aeidon.formats]
    if path.endswith(tuple(extensions)):
        path = path[:path.rfind(".")]
    return "".join((path, format.extension))

def shell_quote(path):
    """Quote and escape `path` for shell use."""
    if sys.platform != "win32":
        # Windows filenames can contain backslashes only as
        # directory separators and cannot contain double quotes.
        path = path.replace("\\", "\\\\")
        path = path.replace('"', '\\"')
    return '"{}"'.format(path)

@contextlib.contextmanager
def silent(*exceptions, tb=False):
    """Try to execute body, ignoring `exceptions`."""
    try:
        yield
    except exceptions:
        if tb: traceback.print_exc()

def start_process(command, **kwargs):
    """
    Start `command` as a new background subprocess.

    `command` and `kwargs` are passed to :class:`subprocess.Popen`.
    Raise :exc:`aeidon.ProcessError` if something goes wrong.
    Return :class:`subprocess.Popen` instance.
    """
    # Use no environment on Windows due to a subprocess bug.
    # https://bugzilla.gnome.org/show_bug.cgi?id=605805
    env = os.environ.copy() if sys.platform != "win32" else None
    try:
        return subprocess.Popen(command,
                                shell=(sys.platform != "win32"),
                                cwd=os.getcwd(),
                                env=env,
                                universal_newlines=True,
                                **kwargs)

    except OSError as error:
        raise aeidon.ProcessError(str(error.args))

def title_to_lower_case(title_name):
    """Convert title case name to lower case with underscores."""
    lower_name = ""
    for char in title_name:
        if char.isupper() and lower_name:
            lower_name += "_"
        lower_name += char.lower()
    return lower_name

def uri_to_path(uri):
    """Convert `uri` to local filepath."""
    uri = urllib.parse.unquote(uri)
    if sys.platform == "win32":
        path = urllib.parse.urlsplit(uri)[2]
        while path.startswith("/"):
            path = path[1:]
        return path.replace("/", "\\")
    return urllib.parse.urlsplit(uri)[2]

def write(path, text, encoding=None, fallback="utf_8", quiet=False):
    """
    Write `text` to file at `path`.

    `fallback` should be ``None`` to try only `encoding`.
    Raise :exc:`IOError` if writing fails.
    Raise :exc:`UnicodeError` if encoding fails.
    """
    if not os.path.isdir(os.path.dirname(path)):
        makedirs(os.path.dirname(path))
    encoding = encoding or get_default_encoding()
    try:
        with open(path, "w", encoding=encoding) as f:
            return f.write(text)
    except IOError:
        if not quiet:
            print_write_io(sys.exc_info(), path)
        raise # IOError
    except UnicodeError:
        if not fallback in (encoding, None, ""):
            return write(path, text, fallback, None, quiet)
        if not quiet:
            print_write_unicode(sys.exc_info(), path, encoding)
        raise # UnicodeError

def writelines(path, lines, encoding=None, fallback="utf_8", quiet=False):
    """
    Write `lines` of text to file at `path`.

    `fallback` should be ``None`` to try only `encoding`.
    Raise :exc:`IOError` if writing fails.
    Raise :exc:`UnicodeError` if encoding fails.
    """
    text = "\n".join(lines) + "\n"
    return write(path, text, encoding, fallback, quiet)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialogs for selecting character encodings."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("EncodingDialog", "MenuEncodingDialog")


class EncodingDialog(Gtk.Dialog):

    """Dialog for selecting a character encoding."""

    def __init__(self, parent):
        """Initialize an :class:`EncodingDialog` instance."""
        GObject.GObject.__init__(self, use_header_bar=True)
        self._tree_view = Gtk.TreeView()
        self._init_dialog(parent)
        self._init_tree_view()
        gaupol.util.scale_to_content(self._tree_view,
                                     min_nchar=50,
                                     max_nchar=100,
                                     min_nlines=10,
                                     max_nlines=18)

    def get_encoding(self):
        """Return the selected encoding or ``None``."""
        selection = self._tree_view.get_selection()
        store, itr = selection.get_selected()
        if itr is None: return
        return store.get_value(itr, 0)

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_OK"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)
        self.set_title(_("Character Encodings"))

    def _init_tree_view(self):
        """Initialize the character encoding tree view."""
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_shadow_type(Gtk.ShadowType.NONE)
        scroller.add(self._tree_view)
        box = self.get_content_area()
        gaupol.util.pack_start_expand(box, scroller)
        box.show_all()
        selection = self._tree_view.get_selection()
        selection.set_mode(Gtk.SelectionMode.SINGLE)
        store = Gtk.ListStore(str, str, str)
        for item in aeidon.encodings.get_valid():
            store.append((item[0], item[2], item[1]))
        store.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        self._tree_view.set_model(store)
        column = Gtk.TreeViewColumn(_("Description"),
                                    Gtk.CellRendererText(),
                                    text=1)

        column.set_clickable(True)
        column.set_sort_column_id(1)
        self._tree_view.append_column(column)
        column = Gtk.TreeViewColumn(_("Encoding"),
                                    Gtk.CellRendererText(),
                                    text=2)

        column.set_clickable(True)
        column.set_sort_column_id(2)
        self._tree_view.append_column(column)
        aeidon.util.connect(self, "_tree_view", "row-activated")

    def _on_tree_view_row_activated(self, *args):
        """Send response to select activated character encoding."""
        self.response(Gtk.ResponseType.OK)


class MenuEncodingDialog(EncodingDialog):

    """Dialog for selecting character encodings."""

    def get_visible_encodings(self):
        """Return encodings chosen to be visible."""
        store = self._tree_view.get_model()
        return [store[i][0] for i in range(len(store)) if store[i][3]]

    def _init_tree_view(self):
        """Initialize the character encoding tree view."""
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_shadow_type(Gtk.ShadowType.NONE)
        scroller.add(self._tree_view)
        box = self.get_content_area()
        gaupol.util.pack_start_expand(box, scroller)
        box.show_all()
        selection = self._tree_view.get_selection()
        selection.set_mode(Gtk.SelectionMode.SINGLE)
        store = Gtk.ListStore(str, str, str, bool)
        visible = gaupol.conf.encoding.visible
        for item in aeidon.encodings.get_valid():
            store.append((item[0], item[2], item[1], item[0] in visible))
        store.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        self._tree_view.set_model(store)
        column = Gtk.TreeViewColumn(_("Description"),
                                    Gtk.CellRendererText(),
                                    text=1)

        column.set_clickable(True)
        column.set_sort_column_id(1)
        self._tree_view.append_column(column)
        column = Gtk.TreeViewColumn(_("Encoding"),
                                    Gtk.CellRendererText(),
                                    text=2)

        column.set_clickable(True)
        column.set_sort_column_id(2)
        self._tree_view.append_column(column)
        renderer = Gtk.CellRendererToggle()
        renderer.connect("toggled", self._on_tree_view_cell_toggled)
        column = Gtk.TreeViewColumn(_("Show in Menu"),
                                    renderer,
                                    active=3)

        column.set_sort_column_id(3)
        self._tree_view.append_column(column)
        aeidon.util.connect(self, "_tree_view", "row-activated")

    def _on_tree_view_cell_toggled(self, renderer, path):
        """Toggle the value of the "Show in Menu" column."""
        store = self._tree_view.get_model()
        store[path][3] = not store[path][3]



# -*- coding: utf-8 -*-

# Copyright (C) 2010 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Reading, writing and storing configurations."""

import aeidon
import copy
import gaupol
import json
import os
import sys

__all__ = ("ConfigurationStore",)


CONFIG_DEFAULTS = {
    "application_window": {
        "layout": gaupol.orientation.VERTICAL,
        "maximized": False,
        "position": [0, 0],
        "show_main_toolbar": True,
        "size": [700, 433],
        "toolbar_style": gaupol.toolbar_styles.ICONS,
    },
    "capitalization": {
        "country": "",
        "language": "",
        "script": "Latn",
    },
    "common_error": {
        "classes": ["Human", "OCR"],
        "country": "",
        "language": "",
        "script": "Latn",
    },
    "duration_adjust": {
        "gap": 0.0,
        "lengthen": True,
        "maximum": 6.0,
        "minimum": 1.5,
        "speed": 15.0,
        "shorten": False,
        "target": gaupol.targets.CURRENT,
        "use_minimum": True,
        "use_gap": True,
        "use_maximum": False,
    },
    "general": {
        "dark_theme": False,
        "diff_color_change": "#ffff0033",
        "diff_color_delete": "#ff555533",
        "diff_color_insert": "#00ff0033",
        "version": None,
    },
    "editor": {
        "custom_font": ("Consolas 9" if sys.platform == "win32"
                        else "monospace"),

        "field_order": [
            gaupol.fields.NUMBER,
            gaupol.fields.START,
            gaupol.fields.END,
            gaupol.fields.DURATION,
            gaupol.fields.MAIN_TEXT,
            gaupol.fields.TRAN_TEXT,
        ],
        "framerate": aeidon.framerates.FPS_23_976,
        "length_unit": gaupol.length_units.EM,
        "mode": aeidon.modes.TIME,
        "show_lengths_cell": True,
        "show_lengths_edit": True,
        "stretch_length": 0.05,
        "use_custom_font": True,
        "use_zebra_stripes": True,
        "visible_fields": [
            gaupol.fields.NUMBER,
            gaupol.fields.START,
            gaupol.fields.END,
            gaupol.fields.DURATION,
            gaupol.fields.MAIN_TEXT,
        ],
    },
    "encoding": {
        "fallback": ["utf_8", "cp1252"],
        "try_auto": True,
        "try_locale": True,
        "visible": ["utf_8", "cp1252"],
    },
    "extensions": {
        "active": [],
    },
    "file": {
        "align_method": aeidon.align_methods.POSITION,
        "directory": "",
        "encoding": "utf_8",
        "format": aeidon.formats.SUBRIP,
        "newline": aeidon.util.get_default_newline(),
    },
    "framerate_convert": {
        "target": gaupol.targets.CURRENT,
    },
    "hearing_impaired": {
        "country": "",
        "language": "",
        "script": "Latn",
    },
    "join_split_words": {
        "join": True,
        "split": False,
    },
    "line_break": {
        "country": "",
        "language": "",
        "length_unit": gaupol.length_units.EM,
        "max_length": 24,
        "max_lines": 3,
        "script": "Latn",
        "skip_max_length": 24,
        "skip_max_lines": 3,
        "use_skip_max_length": True,
        "use_skip_max_lines": True,
    },
    "position_shift": {
        "target": gaupol.targets.CURRENT,
    },
    "position_transform": {
        "target": gaupol.targets.CURRENT,
    },
    "preview": {
        "custom_command": "",
        "force_utf_8": True,
        # mpv is the only player that supports precise seek, others need a longer offset.
        "offset": 1.0 if gaupol.util.get_default_player() == aeidon.players.MPV else 5.0,
        "player": gaupol.util.get_default_player(),
        "use_custom_command": False,
    },
    "recent": {
        "show_not_found": False,
    },
    "search": {
        "fields": [gaupol.fields.MAIN_TEXT],
        "ignore_case": True,
        "regex": False,
        "target": gaupol.targets.CURRENT,
    },
    "spell_check": {
        "field": gaupol.fields.MAIN_TEXT,
        "inline": False,
        "language": "en",
        "size": [500, 309],
        "target": gaupol.targets.CURRENT,
    },
    "subtitle_insert": {
        "above": False,
    },
    "text_assistant": {
        "field": gaupol.fields.MAIN_TEXT,
        "maximized": False,
        "pages": [],
        "remove_blank": True,
        "size": [700, 433],
        "target": gaupol.targets.CURRENT,
    },
    "video_player": {
        "autoplay": True,
        "context_length": 1.0,
        "line_alignment": "center",
        "seek_length": 30.0,
        "subtitle_alpha": 1.0,
        "subtitle_background": True,
        "subtitle_color": "#ffffff",
        "subtitle_font": "Sans 18",
        "subtitle_position_horizontal": "center",
        "subtitle_position_vertical": "bottom",
        "time_alpha": 1.0,
        "time_background": True,
        "time_color": "#ffffff",
        "time_font": "Monospace 14",
        "time_position_horizontal": "right",
        "time_position_vertical": "top",
        "volume": None,
    },
}

CONFIG_ENUMS = {
    "application_window": {
        "layout": gaupol.orientation,
        "toolbar_style": gaupol.toolbar_styles,
    },
    "duration_adjust": {
        "target": gaupol.targets,
    },
    "editor": {
        "field_order": gaupol.fields,
        "framerate": aeidon.framerates,
        "length_unit": gaupol.length_units,
        "mode": aeidon.modes,
        "visible_fields": gaupol.fields,
    },
    "extensions": {
    },
    "file": {
        "align_method": aeidon.align_methods,
        "format": aeidon.formats,
        "newline": aeidon.newlines,
    },
    "framerate_convert": {
        "target": gaupol.targets,
    },
    "line_break": {
        "length_unit": gaupol.length_units,
    },
    "position_shift": {
        "target": gaupol.targets,
    },
    "position_transform": {
        "target": gaupol.targets,
    },
    "preview": {
        "player": aeidon.players,
    },
    "search": {
        "fields": gaupol.fields,
        "target": gaupol.targets,
    },
    "spell_check": {
        "field": gaupol.fields,
        "target": gaupol.targets,
    },
    "text_assistant": {
        "field": gaupol.fields,
        "target": gaupol.targets,
    },
}


class EnumDecoder(json.JSONDecoder):

    """JSON decoder for enumerations of :mod:`aeidon` and :mod:`gaupol`."""

    def __init__(self, *args, **kwargs):
        """
        Initialize an :class:`EnumDecoder` instance.

        `kwargs` should contain an "enum" key with a value of ``None``
        if not decoding an enumeration or the corresponding
        :class:`aeidon.Enumeration` instance if decoding an enumeration.
        """
        self.enum = kwargs["enum"]
        del kwargs["enum"]
        json.JSONDecoder.__init__(self, *args, **kwargs)

    def decode(self, string):
        """Return Python object matching JSON `string`."""
        if self.enum is None:
            return json.JSONDecoder.decode(self, string)
        if string.startswith("[") and string.endswith("]"):
            return [getattr(self.enum, x.strip())
                    for x in string[1:-1].split(",")]
        return getattr(self.enum, string)


class EnumEncoder(json.JSONEncoder):

    """JSON encoder for enumerations of :mod:`aeidon` and :mod:`gaupol`."""

    def encode(self, obj):
        """Return JSON string matching `obj`."""
        if isinstance(obj, aeidon.EnumerationItem):
            return str(obj)
        if isinstance(obj, (list, tuple)) and obj:
            if isinstance(obj[0], aeidon.EnumerationItem):
                return "[{}]".format(", ".join(map(str, obj)))
        return json.JSONEncoder.encode(self, obj)


class ConfigurationStore(gaupol.AttributeDictionary):

    """
    Reading, writing and storing configurations.

    :cvar _defaults: Dictionary of default values of options
    :cvar _enums: Dictionary of :class:`aeidon.Enumeration` instances
    :ivar path: Path to user's local configuration file
    """

    _defaults = copy.deepcopy(CONFIG_DEFAULTS)
    _enums = copy.deepcopy(CONFIG_ENUMS)

    def __init__(self):
        """Initialize a :class:`ConfigurationStore` instance."""
        root = copy.deepcopy(self._defaults)
        gaupol.AttributeDictionary.__init__(self, root)
        self.path = None

    def __setattr__(self, name, value):
        """Avoid accidentally setting a tuple."""
        if isinstance(value, tuple):
            value = list(value)
        return gaupol.AttributeDictionary.__setattr__(self, name, value)

    def connect_notify(self, sections, option, obj, *args):
        """Connect `option`'s notify signal to `obj`'s callback method."""
        if isinstance(sections, str):
            sections = (sections,)
        container = self
        for section in sections:
            container = getattr(container, section)
        signal = "notify::{}".format(option)
        method_name = "_on_conf_{}_{}".format(
            "_".join(sections),
            signal.replace("::", "_"))
        if not hasattr(obj, method_name):
            method_name = method_name[1:]
        method = getattr(obj, method_name)
        container.connect(signal, method, *args)

    def disconnect_notify(self, sections, option, obj):
        """Disconnect `option`'s notify signal from `obj`'s callback method."""
        if isinstance(sections, str):
            sections = (sections,)
        container = self
        for section in sections:
            container = getattr(container, section)
        signal = "notify::{}".format(option)
        method_name = "_on_conf_{}_{}".format(
            "_".join(sections),
            signal.replace("::", "_"))
        if not hasattr(obj, method_name):
            method_name = method_name[1:]
        method = getattr(obj, method_name)
        container.disconnect(signal, method)

    def _flatten(self, values):
        """Return a flattened version of `values` dictionary."""
        def flatten(deep, parent):
            flat_dict = {parent: {}}
            deep = copy.deepcopy(deep)
            for key, value in deep.items():
                if isinstance(value, dict):
                    key = "::".join((parent, key))
                    value = flatten(value, key)
                    flat_dict.update(value)
                else: # Non-dictionary key.
                    flat_dict[parent][key] = value
            if not flat_dict[parent]:
                del flat_dict[parent]
            return flat_dict
        final_dict = {}
        values = copy.deepcopy(values)
        for key, value in values.items():
            if isinstance(value, dict):
                final_dict.update(flatten(value, key))
            else: # Non-dictionary key.
                final_dict[key] = value
        return final_dict

    def query_default(self, sections, option):
        """Return default value of configuration option."""
        if isinstance(sections, str):
            sections = (sections,)
        container = self._defaults
        for section in sections:
            container = container[section]
        return container[option]

    def _read_from_file(self):
        """Read values of configuration options from file."""
        if self.path is None: return
        if not os.path.isfile(self.path): return
        encoding = aeidon.util.get_default_encoding()
        # Ignore possible encoding errors, which are only related to
        # saved file and directory names and not in any way critical.
        with open(self.path, "r", encoding=encoding, errors="replace") as f:
            lines = f.readlines()
        lines = [x.strip() for x in lines]
        lines = [x for x in lines if x and not x.startswith("#")]
        sections, container, enums = None, None, None
        for line in lines:
            if line.startswith("[") and line.endswith("]"):
                sections, container, enums = self._read_section_line(line)
            else: # OPTION = VALUE
                if sections is None or container is None or enums is None:
                    raise ValueError("Missing section header?")
                self._read_option_line(line, sections, container, enums)

    def read_from_file(self):
        """Read values of configuration options from file."""
        try:
            self._read_from_file()
        except Exception as error:
            print("Failed to read configuration file {}: {!s}"
                  .format(self.path, error),
                  file=sys.stderr)

    def _read_option_line(self, line, sections, container, enums):
        """Read option-value line from configuration file."""
        option, value = [x.strip() for x in line.split("=", 1)]
        enum = enums.get(option, None)
        try:
            value = json.loads(value, cls=EnumDecoder, enum=enum)
        except (AttributeError, ValueError):
            return print("Failed to parse value '{}' of option '{}.{}' from configuration file '{}'."
                         .format(value, "::".join(sections), option, self.path),
                         file=sys.stderr)

        if hasattr(container, option) and value is None:
            # Discard a None-value for all non-extension options.
            # None-values are not used for any options, but might
            # accidentally bet set in some corner or error cases.
            # By discarding them here, we ensure a clean start.
            return print("Discarding value '{}' of option '{}.{}' from configuration file '{}'."
                         .format(value, "::".join(sections), option, self.path),
                         file=sys.stderr)

        if not hasattr(container, option):
            # Add attribute if it does not exist in container, which
            # has been initialized from config_defaults. This is needed
            # for extensions, defaults of options of which are not
            # known until later. Any possible obsolete non-extension
            # options created here will be removed before writing in
            # write_to_file.
            container.add_attribute(option, value)
        setattr(container, option, value)

    def _read_section_line(self, line):
        """Read section header line from configuration file."""
        sections = line[1:-1].strip().split("::")
        container = self
        enums = self._enums
        for section in sections:
            if not hasattr(container, section):
                container.add_attribute(section, {})
            container = getattr(container, section)
            enums = enums.get(section, {})
        return sections, container, enums

    def register_extension(self, name, defaults, enums=None):
        """
        Add section and options for extension.

        `name` should be preferrably the module name of the extension and
        it will appear in the section name as ``extensions::name``. `defaults`
        should be a dictionary of default values for options. `enums` should be
        a dictionary of :class:`aeidon.Enumeration` instances corresponding
        to enumeration items that appear in options.
        """
        self._defaults["extensions"].update({name: defaults})
        self.extensions.extend({name: copy.deepcopy(defaults)})
        self._enums["extensions"].update({name: enums or {}})

    def restore_defaults(self):
        """Set all configuration options to their default values."""
        self.update(self._defaults)

    def _write_to_file(self):
        """Write values of configuration options to file."""
        if self.path is None: return
        self.general.version = gaupol.__version__
        aeidon.util.makedirs(os.path.dirname(self.path))
        encoding = aeidon.util.get_default_encoding()
        # Ignore possible encoding errors, which are only related to
        # saved file and directory names and not in any way critical.
        f = open(self.path, "w", encoding=encoding, errors="replace")
        root = self._flatten(self._root)
        defaults = self._flatten(self._defaults)
        for section in sorted(root):
            f.write("\n[{}]\n".format(section))
            for option in sorted(root[section]):
                value = root[section][option]
                json_value = json.dumps(
                    value, cls=EnumEncoder, ensure_ascii=False)
                # Discard removed options, but always keep
                # all options of all extensions.
                if (not section.startswith("extensions::")
                    and (not section in defaults or
                         not option in defaults[section])): continue
                # Write options that remain at their default value
                # (perhaps, but not necessarily unset) as commented out.
                if (section in defaults
                    and option in defaults[section]
                    and value == defaults[section][option]):
                    f.write("# ")
                f.write("{} = {}\n".format(option, json_value))
        f.close()

    def write_to_file(self):
        """Write values of configuration options to file."""
        try:
            self._write_to_file()
        except Exception as error:
            print("Failed to write configuration file {}: {!s}"
                  .format(self.path, error),
                  file=sys.stderr)



# -*- coding: utf-8 -*-

# Copyright (C) 2013 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Floating label that can be overlaid on another widget."""

import gaupol

from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("FloatingLabel",)


class FloatingLabel(Gtk.Box):

    """Floating label that can be overlaid on another widget."""

    # FloatingLabel is inspired by, but not an implementation of,
    # the floating statusbar found in nautilus.
    # https://gitlab.gnome.org/GNOME/nautilus/blob/master/src/nautilus-floating-bar.c

    def __init__(self):
        """Initialize a :class:`FloatingLabel` instance."""
        GObject.GObject.__init__(self, orientation=Gtk.Orientation.HORIZONTAL)
        self._event_box = Gtk.EventBox()
        self._handlers = []
        self._hide_id = None
        self._label = Gtk.Label()
        self.set_halign(Gtk.Align.START)
        self.set_valign(Gtk.Align.END)
        self._init_widgets()

    def _clear_hide_events(self):
        """Disconnect registered hide events."""
        while self._handlers:
            widget, handler_id = self._handlers.pop()
            if widget.handler_is_connected(handler_id):
                widget.handler_disconnect(handler_id)

    def flash_text(self, text, duration=6):
        """Show label with `text` for `duration` seconds."""
        if not text:
            return self._hide()
        self.set_text(text)
        self._hide_id = gaupol.util.delay_add(duration * 1000, self._hide)

    def get_text(self):
        """Return text shown in the label."""
        return self._label.get_text()

    def _hide(self, *args):
        """Hide and disconnect handlers."""
        self.hide()
        self._hide_id = None
        self._clear_hide_events()

    def _init_widgets(self):
        """Initialize widgets contained in the box."""
        style = self._label.get_style_context()
        style.add_class("gaupol-floating-label")
        self._event_box.add(self._label)
        gaupol.util.pack_start(self, self._event_box)
        self._event_box.connect("enter-notify-event", self._hide)

    def register_hide_event(self, widget, signal):
        """Register `widget`'s `signal` as cause to hide the label."""
        handler_id = widget.connect(signal, self._hide)
        self._handlers.append((widget, handler_id))

    def set_text(self, text):
        """Show `text` in the label."""
        if not text:
            return self._hide()
        if self._hide_id is not None:
            GObject.source_remove(self._hide_id)
        self._label.set_text(text)
        self.show()



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous helper methods."""

import aeidon


class UtilityAgent(aeidon.Delegate):

    """Miscellaneous helper methods."""

    @aeidon.deco.export
    def get_all_indices(self):
        """Return a list of all indices of subtitles."""
        return list(range(len(self.subtitles)))

    @aeidon.deco.export
    def get_changed(self, doc):
        """Return the changed value corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return self.main_changed
        if doc == aeidon.documents.TRAN:
            return self.tran_changed
        raise ValueError("Invalid document: {!r}"
                         .format(doc))

    @aeidon.deco.export
    def get_file(self, doc):
        """Return the file corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return self.main_file
        if doc == aeidon.documents.TRAN:
            return self.tran_file
        raise ValueError("Invalid document: {!r}"
                         .format(doc))

    @aeidon.deco.export
    def get_format(self, doc):
        """
        Return format of the file corresponding to `doc`.

        For a translation file that is ``None``, return format of main file.
        If main file is ``None``, return ``None``.
        """
        if doc == aeidon.documents.MAIN:
            if self.main_file is not None:
                return self.main_file.format
            return None
        if doc == aeidon.documents.TRAN:
            if self.tran_file is not None:
                return self.tran_file.format
            return self.get_format(aeidon.documents.MAIN)
        raise ValueError("Invalid document: {!r}"
                         .format(doc))

    @aeidon.deco.export
    def get_liner(self, doc):
        """Return a new :class:`aeidon.Liner` instance."""
        re_tag = self.get_markup_tag_regex(doc)
        clean_func = self.get_markup_clean_func(doc)
        return aeidon.Liner(re_tag, clean_func)

    @aeidon.deco.export
    def get_markup(self, doc):
        """Return `doc`'s markup instance or ``None``."""
        format = self.get_format(doc)
        if format is None: return None
        return aeidon.markups.new(format)

    @aeidon.deco.export
    def get_markup_clean_func(self, doc):
        """Return the function to clean markup or ``None``."""
        format = self.get_format(doc)
        if format is None: return None
        return aeidon.markups.new(format).clean

    @aeidon.deco.export
    def get_markup_tag_regex(self, doc):
        """Return the regular expression for a markup tag or ``None``."""
        format = self.get_format(doc)
        if format is None: return None
        return aeidon.markups.new(format).tag

    @aeidon.deco.export
    def get_mode(self):
        """Return mode of the main file or default."""
        if self.main_file is not None:
            return self.main_file.mode
        return aeidon.modes.TIME

    @aeidon.deco.export
    def get_parser(self, doc):
        """Return a new :class:`aeidon.Parser` instance."""
        re_tag = self.get_markup_tag_regex(doc)
        clean_func = self.get_markup_clean_func(doc)
        return aeidon.Parser(re_tag, clean_func)

    @aeidon.deco.export
    def get_text_length(self, index, doc):
        """Return the amount of characters in text excluding markup."""
        text = self.subtitles[index].get_text(doc)
        re_tag = self.get_markup_tag_regex(doc)
        if re_tag is not None:
            text = re_tag.sub("", text)
        return len(text)

    @aeidon.deco.export
    def get_text_signal(self, doc):
        """Return the ``texts-changed`` signal corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return "main-texts-changed"
        if doc == aeidon.documents.TRAN:
            return "translation-texts-changed"
        raise ValueError("Invalid document: {!r}"
                         .format(doc))

    @aeidon.deco.export
    def new_subtitle(self):
        """Return a new :class:`aeidon.Subtitle` instance."""
        return aeidon.Subtitle(self.get_mode(), self.framerate)

    @aeidon.deco.export
    def new_temp_file(self, doc, encoding=None):
        """
        Return path to a new temporary file with subtitles from `doc`.

        Raise :exc:`IOError` if writing to temporary file fails.
        Raise :exc:`UnicodeError` if encoding temporary file fails.
        """
        file = self.get_file(doc)
        if file is None and doc == aeidon.documents.TRAN:
            # For an unsaved translation document,
            # fall back to main document's properties.
            file = self.get_file(aeidon.documents.MAIN)
        if file is not None:
            path = aeidon.temp.create(file.format.extension)
            encoding = encoding or file.encoding
            temp_file = aeidon.files.new(file.format, path, encoding)
            temp_file.copy_from(file)
        else:
            # If no saved document to pull properties from,
            # fall back to SubRip format and UTF-8 encoding.
            format = aeidon.formats.SUBRIP
            path = aeidon.temp.create(format.extension)
            encoding = encoding or "utf_8"
            temp_file = aeidon.files.new(format, path, encoding)
        self.save(doc, temp_file, keep_changes=False)
        return temp_file.path



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Subtitle files of all formats."""

import aeidon

aeidon.util.install_module("files", lambda: None)

from .ssa        import SubStationAlpha
from .ass        import AdvSubStationAlpha
from .lrc        import LRC
from .microdvd   import MicroDVD
from .mpl2       import MPL2
from .subrip     import SubRip
from .subviewer2 import SubViewer2
from .tmplayer   import TMPlayer
from .webvtt     import WebVTT

__all__ = [
    "SubStationAlpha",
    "AdvSubStationAlpha",
    "LRC",
    "MicroDVD",
    "MPL2",
    "SubRip",
    "SubViewer2",
    "TMPlayer",
    "WebVTT",
]

def add(cls):
    """Add a new :class:`aeidon.SubtitleFile` class."""
    globals()[cls.__name__] = cls
    __all__.append(cls.__name__)

def new(format, path, encoding, newline=None):
    """Return a new :class:`aeidon.SubtitleFile` instance given `format`."""
    for cls in map(eval, __all__):
        if cls.format == format:
            return cls(path, encoding, newline)
    raise ValueError("Format {!r} not found"
                     .format(format))



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Managing regular expression substitutions for subtitle texts."""

import aeidon
import os
import re
import xml.etree.ElementTree as ET

__all__ = ("PatternManager",)


class PatternManager:

    """
    Managing regular expression substitutions for subtitle texts.

    :ivar _patterns: Dictionary mapping codes to pattern lists
    :ivar pattern_type: String to indentify what the pattern matches

    :attr:`pattern_type` should be a string with value "line-break",
    "common-error", "capitalization" or "hearing-impaired". Codes are of form
    ``Script[-language-[COUNTRY]]`` using the corresponding ISO codes.
    """
    _re_comment = re.compile(r"^\s*#.*$")

    def __init__(self, pattern_type):
        """Initialize a :class:`PatternManager` instance."""
        self.pattern_type = pattern_type
        self._patterns = {}
        self._read_patterns()

    def _filter_patterns(self, patterns):
        """
        Return `patterns` with name-clashes resolved.

        Patterns with a more specific code replace those with a less specific
        code if they have the same name and the more specific pattern's policy
        is explicitly set to ``Replace`` (instead of the implicit ``Append``).
        Order is maintained so that all patterns with the same name are always
        located in the position of the earliest of such patterns.
        """
        filtered_patterns = []
        for pattern in patterns:
            name = pattern.get_name(False)
            policy = pattern.get_field("Policy")
            last_index = len(filtered_patterns) - 1
            for j, filtered_pattern in enumerate(filtered_patterns):
                if filtered_pattern.get_name() == name:
                    last_index = j
                    if policy == "Replace":
                        filtered_patterns[j] = None
            filtered_patterns.insert(last_index + 1, pattern)
            while None in filtered_patterns:
                filtered_patterns.remove(None)
        return filtered_patterns

    def _get_codes(self, script=None, language=None, country=None):
        """Return a sequence of all codes to be used by arguments."""
        codes = ["Zyyy"]
        if not None in (script,):
            codes.append(script)
        if not None in (script, language):
            codes.append("{}-{}".format(script, language))
        if not None in (script, language, country):
            codes.append("{}-{}-{}".format(script, language, country))
        return tuple(codes)

    def get_countries(self, script, language):
        """Return a sequence of countries for which patterns exist."""
        codes = list(self._patterns.keys())
        start = "{}-{}-".format(script, language)
        codes = [x for x in codes if x.startswith(start)]
        countries = [x.split("-")[2] for x in codes]
        return tuple(aeidon.util.get_unique(countries))

    def get_languages(self, script):
        """Return a sequence of languages for which patterns exist."""
        codes = list(self._patterns.keys())
        start = "{}-".format(script)
        codes = [x for x in codes if x.startswith(start)]
        languages = [x.split("-")[1] for x in codes]
        return tuple(aeidon.util.get_unique(languages))

    def get_patterns(self, script=None, language=None, country=None):
        """Return patterns for `script`, `language` and `country`."""
        patterns = []
        codes = self._get_codes(script, language, country)
        for code in codes:
            for pattern in self._patterns.get(code, []):
                # Skip patterns that define exceptions to their use
                # that match a more speficic group being requested.
                skip = pattern.get_field_list("SkipIn", [])
                if set(skip) & set(codes): continue
                patterns.append(pattern)
        return self._filter_patterns(patterns)

    def get_scripts(self):
        """Return a sequence of scripts for which patterns exist."""
        codes = list(self._patterns.keys())
        while "Zyyy" in codes:
            codes.remove("Zyyy")
        scripts = [x.split("-")[0] for x in codes]
        return tuple(aeidon.util.get_unique(scripts))

    def _read_config_from_directory(self, directory, encoding):
        """Read configurations from files in `directory`."""
        if not os.path.isdir(directory): return
        extension = ".{}.conf".format(self.pattern_type)
        files = os.listdir(directory)
        for name in (x for x in files if x.endswith(extension)):
            path = os.path.join(directory, name)
            self._read_config_from_file(path, encoding)

    def _read_config_from_file(self, path, encoding):
        """Read configurations from file at `path`."""
        if not os.path.isfile(path): return
        basename = os.path.basename(path)
        extension = ".{}.conf".format(self.pattern_type)
        code = basename.replace(extension, "")
        if not code in self._patterns: return
        patterns = self._patterns[code]
        for element in ET.parse(path).findall("pattern"):
            name = element.get("name")
            name = name.replace("&quot;", '"')
            name = name.replace("&amp;", "&")
            enabled = (element.get("enabled") == "true")
            for pattern in patterns:
                if pattern.get_name(localize=False) == name:
                    pattern.enabled = enabled

    def _read_patterns(self):
        """Read all patterns of :attr:`pattern_type` from files."""
        data_dir = os.path.join(aeidon.DATA_DIR, "patterns")
        data_home_dir = os.path.join(aeidon.DATA_HOME_DIR, "patterns")
        config_home_dir = os.path.join(aeidon.CONFIG_HOME_DIR, "patterns")
        encoding = aeidon.util.get_default_encoding()
        self._read_patterns_from_directory(data_dir, "utf_8")
        self._read_patterns_from_directory(data_home_dir, encoding)
        self._read_config_from_directory(data_dir, "utf_8")
        self._read_config_from_directory(config_home_dir, encoding)

    def _read_patterns_from_directory(self, directory, encoding):
        """Read all patterns from files in `directory`."""
        if not os.path.isdir(directory): return
        extension = ".{}".format(self.pattern_type)
        extensions = (extension, "{}.in".format(extension))
        files = [x for x in os.listdir(directory)
                 if x.endswith(extensions)]

        for name in [x for x in files if x.endswith(".in")]:
            # If both untranslated and translated pattern files are found,
            # load patterns only from the translated one.
            if name[:-3] in files:
                files.remove(name)
        for name in files:
            path = os.path.join(directory, name)
            self._read_patterns_from_file(path, encoding)

    def _read_patterns_from_file(self, path, encoding):
        """Read all patterns from file at `path`."""
        if not os.path.isfile(path): return
        basename = os.path.basename(path)
        extension = ".{}".format(self.pattern_type)
        if basename.endswith(".in"):
            extension = ".{}.in".format(self.pattern_type)
        code = basename.replace(extension, "")
        local = path.startswith(aeidon.DATA_HOME_DIR)
        patterns = self._patterns.setdefault(code, [])
        lines = aeidon.util.readlines(path, encoding)
        lines = [self._re_comment.sub("", x) for x in lines]
        lines = [x.strip() for x in lines]
        for line in (x for x in lines if x):
            if line.startswith("["): # [HEADER]
                patterns.append(aeidon.Pattern())
                patterns[-1].local = local
            else: # [_]KEY=VALUE
                name, value = line.split("=", 1)
                # Translatable fields used to be marked with a leading
                # underscore prior to version 1.3. We continue to support that
                # syntax in case users have local pattern files.
                name = name[1:] if name.startswith("_") else name
                # Regular expression patterns and replacements use null
                # character to avoid syntax issues that go against the GKeyFile
                # spec and would be "fixed" by msgfmt when merging translations.
                # https://github.com/otsaloma/gaupol/issues/70
                value = re.sub(r"\\0(?!\d)", "", value)
                patterns[-1].set_field(name, value)

    def save_config(self, script=None, language=None, country=None):
        """Save pattern configurations to files."""
        codes = self._get_codes(script, language, country)
        for code in (x for x in codes if x in self._patterns):
            self._write_config_to_file(code, "utf_8")

    def _write_config_to_file(self, code, encoding):
        """Write configurations of all patterns to file."""
        local_dir = os.path.join(aeidon.CONFIG_HOME_DIR, "patterns")
        basename = "{}.{}.conf".format(code, self.pattern_type)
        path = os.path.join(local_dir, basename)
        lines = ['<?xml version="1.0" encoding="utf-8"?>']
        lines.append('<patterns>')
        written_names = set()
        for pattern in self._patterns[code]:
            name = pattern.get_name(localize=False)
            if name in written_names: continue
            written_names.add(name)
            name = name.replace("&", "&amp;")
            name = name.replace('"', "&quot;")
            enabled = "true" if pattern.enabled else "false"
            lines.append('  <pattern name="{}" enabled="{}"/>'
                         .format(name, enabled))

        lines.append('</patterns>')
        aeidon.util.writelines(path, lines, encoding)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for inserting new subtitles."""

import gaupol

from aeidon.i18n   import _
from gi.repository import Gtk

__all__ = ("InsertDialog",)


class InsertDialog(gaupol.BuilderDialog):

    """Dialog for inserting new subtitles."""

    _widgets = ["amount_spin", "position_combo", "position_label"]

    def __init__(self, parent, application):
        """Initialize an :class:`InsertDialog` instance."""
        gaupol.BuilderDialog.__init__(self, "insert-dialog.ui")
        self.application = application
        self._init_dialog(parent)
        self._init_position_combo()
        self._init_values()

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Insert"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)

    def _init_position_combo(self):
        """Initialize the position combo box."""
        store = Gtk.ListStore(str)
        self._position_combo.set_model(store)
        store.append((_("Above selection"),))
        store.append((_("Below selection"),))
        renderer = Gtk.CellRendererText()
        self._position_combo.pack_start(renderer, expand=True)
        self._position_combo.add_attribute(renderer, "text", 0)

    def _init_values(self):
        """Initialize default values for widgets."""
        self._amount_spin.set_value(1)
        index = 0 if gaupol.conf.subtitle_insert.above else 1
        self._position_combo.set_active(index)
        page = self.application.get_current_page()
        sensitive = bool(page.project.subtitles)
        self._position_combo.set_sensitive(sensitive)
        self._position_label.set_sensitive(sensitive)

    def _insert_subtitles(self, amount, above):
        """Insert `amount` of subtitles to project."""
        page = self.application.get_current_page()
        if page.project.subtitles:
            index = page.view.get_selected_rows()[-1]
            if not above: index += 1
        else: # No subtitles in project.
            index = 0
        indices = list(range(index, index + amount))
        page.project.insert_subtitles(indices)

    def _on_response(self, dialog, response):
        """Save default values and insert subtitles."""
        amount = self._amount_spin.get_value_as_int()
        above = (self._position_combo.get_active() == 0)
        gaupol.conf.subtitle_insert.above = above
        if response == Gtk.ResponseType.OK:
            self._insert_subtitles(amount, above)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Text markup for the MPL2 format."""

import aeidon

from collections import OrderedDict

__all__ = ("MPL2",)


class MPL2(aeidon.markups.MicroDVD):

    """
    Text markup for the MPL2 format.

    MPL2 contains the following markup tags, all of which appear at the
    beginning of the line and affect up to the end of the line. In addition,
    any MicroDVD markup can be used.

     * ``\\...`` (bold)
     * ``/....`` (italic)
     * ``_....`` (underline)
    """

    format = aeidon.formats.MPL2

    def bolden(self, text, bounds=None):
        """Return bolded `text`."""
        return self._style_mpl2(text, "\\", bounds)

    @property
    def italic_tag(self):
        """Regular expression for an italic markup tag."""
        return self._get_regex(r"(^/)|(\{[Yy]:i\})")

    def italicize(self, text, bounds=None):
        """Return italicized `text`."""
        return self._style_mpl2(text, "/", bounds)

    def _main_decode(self, text):
        """Return `text` with decodable markup decoded."""
        text = self._decode_b(text, r"<\\>(.*?)</\\>", 1)
        text = self._decode_i(text, r"</>(.*?)<//>", 1)
        text = self._decode_u(text, r"<_>(.*?)</_>", 1)
        return aeidon.markups.MicroDVD._main_decode(self, text)

    def _pre_decode(self, text):
        """Return `text` with markup prepared for decoding."""
        text = self._pre_decode_identify(text)
        return aeidon.markups.MicroDVD._pre_decode(self, text)

    def _pre_decode_identify(self, text):
        """
        Return `text` with all tags identified and closed.

        ``\\``, ``/`` and ``_`` characters at the beginnings of lines are
        identified as tags and replaced with artificial tags ``<\\>``, ``</>``
        and ``<_>``. Closing tags are added to the ends of lines as artificial
        tags ``</\\>``, ``<//>`` and ``</_>``.
        """
        lines = text.split("\n")
        re_tag = self._get_regex(r"^([\\/_]+)(.*)$")
        for i, line in enumerate(lines):
            match = re_tag.search(line)
            if match is None: continue
            lines[i] = match.group(2)
            for tag in reversed(OrderedDict.fromkeys(match.group(1))):
                lines[i] = "<{}>{}</{}>".format(tag, lines[i], tag)
        return "\n".join(lines)

    def _style_mpl2(self, text, tag, bounds=None):
        """Return `text` wrapped in markup `tag`."""
        a, z = bounds or (0, len(text))
        prefix = text[:a].split("\n")[-1]
        suffix = text[z:].split("\n")[0]
        re_alpha = self._get_regex(r"\w")
        # Return plain text if bounds does not define an entire line or
        # subtitle and thus cannot be marked without side-effects.
        if re_alpha.search(prefix): return text
        if re_alpha.search(suffix): return text
        styled_text = text[a:z].replace("\n", "\n{}".format(tag))
        return "".join((text[:a], tag, styled_text, text[z:]))

    @property
    def tag(self):
        """Regular expression for any markup tag."""
        return self._get_regex(r"(^[\\/_]+)|(\{[CFSYcfsy]:.*?\})")

    def underline(self, text, bounds=None):
        """Return underlined `text`."""
        return self._style_mpl2(text, "_", bounds)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting subtitle files to open."""

import aeidon
import gaupol
import os

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("OpenDialog",)


class OpenDialog(Gtk.FileChooserDialog, gaupol.FileDialog):

    """Dialog for selecting subtitle files to open."""

    _widgets = ["align_combo", "align_label", "encoding_combo"]

    def __init__(self, parent, title, doc):
        """Initialize an :class:`OpenDialog` instance."""
        GObject.GObject.__init__(self)
        self._use_autodetection = aeidon.util.chardet_available()
        self._init_dialog(parent, title)
        self._init_extra_widget()
        self._init_filters()
        self._init_encoding_combo()
        self._init_align_combo()
        self._init_values(doc)

    def _init_align_combo(self):
        """Initialize the align method combo box."""
        store = Gtk.ListStore(str)
        self._align_combo.set_model(store)
        for align_method in aeidon.align_methods:
            store.append((align_method.label,))
        view = self._align_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._align_combo.pack_start(renderer, expand=True)
        self._align_combo.add_attribute(renderer, "text", 0)

    def _init_dialog(self, parent, title):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Open"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_title(title)
        self.connect("response", self._on_response)
        self.set_action(Gtk.FileChooserAction.OPEN)

    def _init_extra_widget(self):
        """Initialize the extra widget from UI definition file."""
        ui_file_path = os.path.join(aeidon.DATA_DIR, "ui", "open-dialog.ui")
        builder = Gtk.Builder()
        builder.set_translation_domain("gaupol")
        builder.add_from_file(ui_file_path)
        builder.connect_signals(self)
        for name in self._widgets:
            widget = builder.get_object(name)
            setattr(self, "_{}".format(name), widget)
        vbox = gaupol.util.new_vbox(spacing=0)
        main_vbox = builder.get_object("main_vbox")
        main_vbox.get_parent().remove(main_vbox)
        vbox.add(main_vbox)
        vbox.show_all()
        self.set_extra_widget(vbox)

    def _init_values(self, doc):
        """Initialize default values for widgets."""
        self.set_select_multiple(doc == aeidon.documents.MAIN)
        if os.path.isdir(gaupol.conf.file.directory):
            self.set_current_folder(gaupol.conf.file.directory)
        self.set_encoding(gaupol.conf.file.encoding)
        self._align_combo.set_active(gaupol.conf.file.align_method)
        self._align_combo.set_visible(doc == aeidon.documents.TRAN)
        self._align_label.set_visible(doc == aeidon.documents.TRAN)

    def _on_response(self, dialog, response):
        """Save default values for widgets."""
        gaupol.conf.file.encoding = self.get_encoding()
        directory = self.get_current_folder()
        if directory is not None:
            gaupol.conf.file.directory = directory
        index = self._align_combo.get_active()
        gaupol.conf.file.align_method = aeidon.align_methods[index]



# -*- coding: utf-8 -*-

# Copyright (C) 2008 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Containers for additional format-specific subtitle attributes."""


class SubRip:

    """
    Subtitle box pixel coordinates for extended SubRip format.

    :ivar x1: Subtitle corner X coordinate in pixels
    :ivar y1: Subtitle corner Y coordinate in pixels
    :ivar x2: Subtitle corner X coordinate in pixels
    :ivar y2: Subtitle corner Y coordinate in pixels
    """

    x1 = 0
    y1 = 0
    x2 = 0
    y2 = 0


class SubStationAlpha:

    """
    Attributes for all versions of Sub Station Alpha formats.

    :ivar marked: 0 for not marked or 1 for marked (default 0)
    :ivar layer: Layer used by collusion detection (default 0)
    :ivar style: Style name (default "Default")
    :ivar name: Name of the character speaking (default "")
    :ivar margin_l: Left margin override (default 0, i.e. default margins)
    :ivar margin_r: Right margin override (default 0, i.e. default margins)
    :ivar margin_v: Bottom margin override (default 0, i.e. default margins)
    :ivar effect: Transition effect (default "", i.e. no effect)
    """

    marked = 0
    layer = 0
    style = "Default"
    name = ""
    margin_l = 0
    margin_r = 0
    margin_v = 0
    effect = ""


class WebVTT:

    """
    Attributes for the WebVTT format.

    :ivar comment: Optional comment string (including "NOTE")
    :ivar id: Optional cue identifier
    :ivar settings: Optional cue settings (position etc.)
    :ivar style: Optional CSS style block (including "STYLE")

    https://developer.mozilla.org/en-US/docs/Web/API/WebVTT_API
    """

    comment = ""
    id = ""
    settings = ""
    style = ""


def new(name):
    """Return a new container instance given the container's `name`."""
    if name == "ssa":
        return SubStationAlpha()
    if name == "subrip":
        return SubRip()
    if name == "webvtt":
        return WebVTT()
    raise ValueError("Invalid name: {!r}"
                     .format(name))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base class for unit test cases."""

import aeidon
import os

__all__ = ("TestCase",)


class TestCase:

    """Base class for unit test cases."""

    def assert_raises(self, exception, function, *args, **kwargs):
        """Assert that calling `function` raises `exception`."""
        try:
            function(*args, **kwargs)
        except exception:
            return
        raise AssertionError(
            "{!r} failed to raise {!r}"
            .format(function, exception))

    def get_sample_text(self, format, name=None):
        """
        Return sample text for subtitle file `format`.

        `name` can be specified if different from ``format.name.lower()``.
        This is useful for variants of a format, e.g. "subrip-extended".
        """
        name = name or format.name.lower()
        basename = "".join((name, format.extension))
        path = os.path.join(aeidon.DATA_DIR, "samples", basename)
        with open(path, "r", encoding="ascii") as f:
            return f.read().strip()

    def get_spell_check_language(self, language):
        """Return spell-check language to use in unit tests."""
        # Allow any close-enough variant as different systems will have
        # different spell-check engines and dictionaries installed.
        for candidate in aeidon.SpellChecker.list_languages():
            if candidate.startswith(language):
                return candidate
        raise Exception("Spell-check dictionary {}* not found".format(language))

    def new_microdvd_file(self):
        """Return path to a new temporary MicroDVD file."""
        return self.new_temp_file(aeidon.formats.MICRODVD)

    def new_project(self):
        """Return a new project with both main and translation files."""
        project = aeidon.Project()
        project.open_main(self.new_subrip_file(), "ascii")
        project.open_translation(self.new_microdvd_file(), "ascii")
        return project

    def new_subrip_file(self):
        """Return path to a new temporary SubRip file."""
        return self.new_temp_file(aeidon.formats.SUBRIP)

    def new_temp_file(self, format, name=None):
        """Return path to a new temporary subtitle file."""
        text = self.get_sample_text(format, name)
        path = aeidon.temp.create(format.extension)
        with open(path, "w", encoding="ascii") as f:
            f.write(text)
        return path

    def setUp(self):
        """Compatibility alias for :meth:`setup_method`."""
        self.setup_method(None)

    def setup_method(self, method):
        """Set state for executing tests in `method`."""
        pass

    def tearDown(self):
        """Compatibility alias for :meth:`teardown_method`."""
        self.teardown_method(None)

    def teardown_method(self, method):
        """Remove state set for executing tests in `method`."""
        pass

    def test___init__(self):
        """Make sure that :meth:`setup_method` is always run."""
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""MicroDVD file."""

import aeidon
import re

__all__ = ("MicroDVD",)


class MicroDVD(aeidon.SubtitleFile):

    """MicroDVD file."""

    format = aeidon.formats.MICRODVD
    mode = aeidon.modes.FRAME
    _re_line = re.compile(r"^\{(-?\d+)\}\{(-?\d+)\}(.*?)$")

    def read(self):
        """
        Read file and return subtitles.

        Raise :exc:`IOError` if reading fails.
        Raise :exc:`UnicodeError` if decoding fails.
        """
        subtitles = []
        for line in self._read_lines():
            match = self._re_line.match(line)
            if match is not None:
                subtitle = self._get_subtitle()
                subtitle.start_frame = int(match.group(1))
                subtitle.end_frame = int(match.group(2))
                subtitle.main_text = match.group(3).replace("|", "\n")
                subtitles.append(subtitle)
            elif line.startswith("{DEFAULT}"):
                self.header = line
        return subtitles

    def write_to_file(self, subtitles, doc, f):
        """
        Write `subtitles` from `doc` to file `f`.

        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        """
        if self.header.strip():
            f.write(self.header + "\n")
        for subtitle in subtitles:
            text = subtitle.get_text(doc).replace("\n", "|")
            f.write(("{{{:d}}}{{{:d}}}{}\n"
                     .format(subtitle.start_frame,
                             subtitle.end_frame,
                             text)))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""File and project actions for :class:`gaupol.Application`."""

import aeidon
import gaupol

class CloseProjectAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "close-project")
        self.accelerators = ["<Control>W"]
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class NewProjectAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "new-project")
        self.accelerators = ["<Control>N"]
        self.action_group = "safe"

class OpenMainFilesAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "open-main-files")
        self.accelerators = ["<Control>O"]
        self.action_group = "safe"

class OpenTranslationFileAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "open-translation-file")
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.main_file is not None)

class QuitAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "quit")
        self.accelerators = ["<Control>Q"]
        self.action_group = "safe"

class SaveMainDocumentAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "save-main-document")
        self.accelerators = ["<Control>S"]
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class SaveMainDocumentAsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "save-main-document-as")
        self.accelerators = ["<Shift><Control>S"]
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class SaveTranslationDocumentAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "save-translation-document")
        self.accelerators = ["<Control>T"]
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class SaveTranslationDocumentAsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "save-translation-document-as")
        self.accelerators = ["<Shift><Control>T"]
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2009 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer for integer data."""

import aeidon
import gaupol
import re

from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("IntegerCellRenderer",)


class IntegerCellRenderer(Gtk.CellRendererText):

    """Cell renderer for integer data."""

    __gtype_name__ = "IntegerCellRenderer"

    def __init__(self):
        """Initialize a :class:`IntegerCellRenderer` instance."""
        GObject.GObject.__init__(self)
        aeidon.util.connect(self, self, "edited")
        aeidon.util.connect(self, self, "editing-started")

    def _on_edited(self, renderer, path, text):
        """Check `text` for validity before sending onwards."""
        if re.match(r"[+-]?\d+", text) is None:
            renderer.stop_emission("edited")
            renderer.stop_editing(True)

    def _on_editing_started(self, renderer, editor, path):
        """Set `editor` to use same font as `self`."""
        gaupol.style.use_font(editor, "custom")



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Setting values of single subtitle fields."""

import aeidon
import bisect

from aeidon.i18n import _


class SetAgent(aeidon.Delegate):

    """Setting values of single subtitle fields."""

    def _move_if_needed(self, index):
        """Move subtitle for correct order and return new index."""
        subtitle = self.subtitles[index]
        subtitles = self.subtitles[:index] + self.subtitles[(index + 1):]
        new_index = bisect.bisect_right(subtitles, subtitle)
        if new_index == index: return new_index
        subtitle = self.subtitles.pop(index)
        self.emit("subtitles-removed", (index,))
        self.subtitles.insert(new_index, subtitle)
        self.emit("subtitles-inserted", (new_index,))
        return new_index

    @aeidon.deco.export
    @aeidon.deco.revertable
    def set_duration(self, index, value, register=-1):
        """
        Set the value of duration position.

        `value` can be time, frame or seconds. Use :func:`aeidon.as_time`,
        :func:`aeidon.as_frame` or :func:`aeidon.as_seconds` if necessary
        to ensure correct type.
        """
        subtitle = self.subtitles[index]
        orig_end = subtitle.end
        subtitle.duration = value
        if subtitle.end == orig_end: return
        action = aeidon.RevertableAction(register=register)
        action.docs = tuple(aeidon.documents)
        action.description = _("Editing position")
        action.revert_function = self.set_end
        action.revert_args = (index, orig_end)
        self.register_action(action)
        self.emit("positions-changed", (index,))

    @aeidon.deco.export
    @aeidon.deco.revertable
    def set_end(self, index, value, register=-1):
        """
        Set the value of end position.

        `value` can be time, frame or seconds. Use :func:`aeidon.as_time`,
        :func:`aeidon.as_frame` or :func:`aeidon.as_seconds` if necessary
        to ensure correct type.
        """
        subtitle = self.subtitles[index]
        orig_value = subtitle.end
        subtitle.end = value
        if subtitle.end == orig_value: return
        action = aeidon.RevertableAction(register=register)
        action.docs = tuple(aeidon.documents)
        action.description = _("Editing position")
        action.revert_function = self.set_end
        action.revert_args = (index, orig_value)
        self.register_action(action)
        self.emit("positions-changed", (index,))

    @aeidon.deco.export
    @aeidon.deco.revertable
    def set_main_text(self, index, value, register=-1):
        """Set the value of main document's text."""
        return self.set_text(index,
                             aeidon.documents.MAIN,
                             value,
                             register=register)

    @aeidon.deco.export
    @aeidon.deco.revertable
    @aeidon.deco.notify_frozen
    def set_start(self, index, value, register=-1):
        """
        Set the value of start position.

        `value` can be time, frame or seconds. Use :func:`aeidon.as_time`,
        :func:`aeidon.as_frame` or :func:`aeidon.as_seconds` if necessary
        to ensure correct type.
        """
        subtitle = self.subtitles[index]
        orig_value = subtitle.start
        subtitle.start = value
        if subtitle.start == orig_value: return
        index = self._move_if_needed(index)
        action = aeidon.RevertableAction(register=register)
        action.docs = tuple(aeidon.documents)
        action.description = _("Editing position")
        action.revert_function = self.set_start
        action.revert_args = (index, orig_value)
        self.register_action(action)
        self.emit("positions-changed", (index,))

    @aeidon.deco.export
    @aeidon.deco.revertable
    def set_text(self, index, doc, value, register=-1):
        """Set the value of `doc`'s text."""
        subtitle = self.subtitles[index]
        orig_value = subtitle.get_text(doc)
        if value == orig_value: return
        subtitle.set_text(doc, value)
        action = aeidon.RevertableAction(register=register)
        action.docs = (doc,)
        action.description = _("Editing text")
        action.revert_function = self.set_text
        action.revert_args = (index, doc, orig_value)
        self.register_action(action)
        signal = self.get_text_signal(doc)
        self.emit(signal, (index,))

    @aeidon.deco.export
    @aeidon.deco.revertable
    def set_translation_text(self, index, value, register=-1):
        """Set the value of translation document's text."""
        return self.set_text(index,
                             aeidon.documents.TRAN,
                             value,
                             register=register)



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Internationalization functions."""

import aeidon
import gettext
import locale

_translation = gettext.NullTranslations()


def bind(localedir=aeidon.LOCALE_DIR):
    """Bind translation domains and initialize gettext."""
    with aeidon.util.silent(Exception):
        # Set locale to the user's default setting.
        # Might fail on misconfigured systems.
        locale.setlocale(locale.LC_ALL, "")
    # Make translations available to the gettext module.
    gettext.bindtextdomain("gaupol", localedir)
    gettext.textdomain("gaupol")
    with aeidon.util.silent(Exception):
        # Make translations available to GTK as well.
        # Not available on all platforms.
        locale.bindtextdomain("gaupol", localedir)
        locale.textdomain("gaupol")
    globals()["_translation"] = gettext.translation(
        "gaupol", localedir=localedir, fallback=True)

def _(message):
    """Return the localized translation of `message`."""
    return _translation.gettext(message)

def d_(domain, message):
    """Return the localized translation of `message` from `domain`."""
    return gettext.dgettext(domain, message)

def n_(singular, plural, n):
    """Return the localized translation of `singular` or `plural`."""
    return _translation.ngettext(singular, plural, n)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting a subtitle file to append to project."""

import aeidon
import gaupol

from aeidon.i18n import _

__all__ = ("AppendDialog",)


class AppendDialog(gaupol.OpenDialog):

    """Dialog for selecting a subtitle file to append to project."""

    def __init__(self, parent):
        """Initialize an :class:`AppendDialog` instance."""
        doc = aeidon.documents.MAIN
        gaupol.OpenDialog.__init__(self, parent, _("Append File"), doc)
        self.set_select_multiple(False)



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base class for single-instance classes."""

__all__ = ("Singleton",)


class Singleton:

    """Base class for single-instance classes."""

    def __new__(cls, *args, **kwargs):
        """Return possibly existing instance."""
        if not "_instance" in cls.__dict__:
            cls._instance = object.__new__(cls, *args, **kwargs)
        return cls._instance



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for displaying credits and information."""

import gaupol

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("AboutDialog",)


class AboutDialog(Gtk.AboutDialog):

    """Dialog for displaying credits and information."""

    def __init__(self, parent):
        """Initialize an :class:`AboutDialog` instance."""
        GObject.GObject.__init__(self)
        self.set_title(_("About Gaupol"))
        self.set_transient_for(parent)
        self.set_artists(("Osmo Salomaa <otsaloma@iki.fi>",))
        self.set_authors(("Osmo Salomaa <otsaloma@iki.fi>",))
        self.set_comments(_("Subtitle editor"))
        self.set_copyright("Copyright © 2005–2021 Osmo Salomaa")
        self.set_license_type(Gtk.License.GPL_3_0)
        self.set_logo_icon_name("io.otsaloma.gaupol")
        self.set_program_name("Gaupol")
        # TRANSLATORS: This is a special message that shouldn't be translated
        # literally. It is used in the about dialog to give credits to the
        # translators. Thus, you should translate it to your name and email
        # address. You can also include other translators who have contributed
        # to this translation; in that case, please write them on separate
        # lines seperated by newlines (\n).
        self.set_translator_credits(_("translator-credits"))
        self.set_version(gaupol.__version__)
        self.set_website(gaupol.HOMEPAGE_URL)
        self.set_website_label(_("Gaupol Website"))
        self.set_wrap_license(True)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tool actions for :class:`gaupol.Application`."""

import aeidon
import gaupol

class AdjustDurationsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "adjust-durations")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class AppendFileAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "append-file")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(len(page.project.subtitles) > 0)

class CheckSpellingAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "check-spelling")
        self.accelerators = ["F7"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(aeidon.SpellChecker.available())
        aeidon.util.affirm(gaupol.conf.spell_check.language)

class ConfigureSpellCheckAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "configure-spell-check")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(aeidon.SpellChecker.available())

class ConvertFramerateAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "convert-framerate")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.main_file is not None)

class CorrectTextsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "correct-texts")
        self.accelerators = ["F8"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class PreviewAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "preview")
        self.accelerators = ["F5"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.video_path is not None)
        if gaupol.conf.preview.use_custom_command:
            aeidon.util.affirm(gaupol.conf.preview.custom_command)

class SelectVideoFileAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "select-video-file")
        self.action_group = "safe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.main_file is not None)

class ShiftPositionsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "shift-positions")
        self.accelerators = ["H"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class SplitProjectAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "split-project")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(len(page.project.subtitles) > 1)

class TransformPositionsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "transform-positions")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(len(page.project.subtitles) > 1)

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""GTK user interface controller for :class:`aeidon.Project`."""

import aeidon
import gaupol
import itertools
import sys

from aeidon.i18n   import _
from gi.repository import Gdk
from gi.repository import Gtk

__all__ = ("Application",)


class ApplicationMeta(type):

    """
    Application metaclass with delegated methods added.

    Public methods are added to the class dictionary during :meth:`__new__`
    in order to fool Sphinx (and perhaps other API documentation generators)
    into thinking that the resulting instantiated class actually contains those
    methods, which it does not since the methods are removed during
    :meth:`Application.__init__`.
    """

    def __new__(meta, class_name, bases, dic):
        new_dict = dic.copy()
        for agent_class_name in gaupol.agents.__all__:
            agent_class = getattr(gaupol.agents, agent_class_name)
            def is_delegate_method(name):
                value = getattr(agent_class, name)
                return (callable(value) and
                        hasattr(value, "export") and
                        value.export is True)

            attr_names = list(filter(is_delegate_method, dir(agent_class)))
            for attr_name in attr_names:
                new_dict[attr_name] = getattr(agent_class, attr_name)
        return type.__new__(meta, class_name, bases, new_dict)


class Application(aeidon.Observable, metaclass=ApplicationMeta):

    """
    GTK user interface controller for :class:`aeidon.Project`.

    :ivar clipboard: Instance of :class:`aeidon.Clipboard` used
    :ivar counter: Iterator used for naming unsaved documents
    :ivar _delegations: Dictionary mapping method names to agent methods
    :ivar extension_manager: Instance of :class:`gaupol.ExtensionManager` used
    :ivar main_toolbar: A :class:`Gtk.Toolbar` shown below the menubar
    :ivar notebook: A :class:`Gtk.Notebook` used to hold multiple projects
    :ivar notebook_separator: A :class:`Gtk.Separator` above the notebook
    :ivar open_button: The open button on the main toolbar
    :ivar pages: List of :class:`gaupol.Page` currently open
    :ivar paned: A :class:`Gtk.Paned` to hold player and subtitles
    :ivar pattern: Last used search pattern or blank if not used
    :ivar play_button: The play/pause button on the video player toolbar
    :ivar player: A :class:`gaupol.VideoPlayer` instance or ``None``
    :ivar player_box: Box containing video player etc.
    :ivar player_toolbar: A :class:`Gtk.Toolbar` for video player actions
    :ivar recent_manager: Instance of :class:`Gtk.RecentManager` used
    :ivar redo_button: The redo button on the main toolbar
    :ivar replacement: Last used search replacement or blank if not used
    :ivar seekbar: Video player seekbar (a :class:`Gtk.Scale` instance)
    :ivar statuslabel: Instance of :class:`gaupol.FloatingLabel` used
    :ivar undo_button: The undo button on the main toolbar
    :ivar volume_button: A :class:`Gtk.VolumeButton` in the player toolbar
    :ivar window: A :class:`Gtk.ApplicationWindow` used to hold all the widgets
    :ivar x_clipboard: A :class:`Gtk.Clipboard` used for desktop-wide copying

    Signals and their arguments for callback functions:
     * ``init-done``: application
     * ``page-added``: application, page
     * ``page-changed``: application, page
     * ``page-closed``: application, page
     * ``page-saved``: application, page
     * ``page-switched``: application, page
     * ``pages-reordered``: application, page, number
     * ``quit``: application
    """

    signals = (
        "init-done",
        "page-added",
        "page-changed",
        "page-closed",
        "page-saved",
        "page-switched",
        "pages-reordered",
        "quit",
    )

    def __init__(self):
        """Initialize an :class:`Application` instance."""
        aeidon.Observable.__init__(self)
        self.clipboard = aeidon.Clipboard()
        self.counter = itertools.count(1)
        self._delegations = {}
        self.extension_manager = gaupol.ExtensionManager(self)
        self.main_toolbar = None
        self.notebook = None
        self.notebook_separator = None
        self.open_button = None
        self.pages = []
        self.pattern = ""
        self.play_button = None
        self.player = None
        self.player_box = None
        self.player_toolbar = None
        self.recent_manager = Gtk.RecentManager.get_default()
        self.redo_button = None
        self.replacement = ""
        self.seekbar = None
        self.statuslabel = None
        self.undo_button = None
        self.volume_button = None
        self.window = None
        self.x_clipboard = None
        self._init_delegations()
        self._init_gui()
        self.extension_manager.find_extensions()
        self.extension_manager.setup_extensions()
        self.update_gui()
        self.window.show()
        self.emit("init-done")

    def __getattr__(self, name):
        """Return method delegated to an agent."""
        try:
            return self._delegations[name]
        except KeyError:
            raise AttributeError

    def __setattr__(self, name, value):
        """Set value of attribute `name`."""
        return aeidon.Observable.__setattr__(self, name, value)

    def _init_actions(self):
        """Initialize user-activatable actions."""
        for name in gaupol.actions.__all__:
            action = getattr(gaupol.actions, name)()
            name = "win.{}".format(action.props.name)
            if hasattr(gaupol, "appman"):
                gaupol.appman.set_accels_for_action(
                    name, action.accelerators)
            callback = "_on_{}_activate".format(
                action.props.name.replace("-", "_"))
            action.connect("activate", getattr(self, callback))
            self.window.add_action(action)

    def _init_delegations(self):
        """Initialize the delegation mappings."""
        for agent_class_name in gaupol.agents.__all__:
            agent = getattr(gaupol.agents, agent_class_name)(self)
            def is_delegate_method(name):
                value = getattr(agent, name)
                return (callable(value) and
                        hasattr(value, "export") and
                        value.export is True)

            attr_names = list(filter(is_delegate_method, dir(agent)))
            for attr_name in attr_names:
                attr_value = getattr(agent, attr_name)
                if attr_name in self._delegations:
                    raise ValueError("Multiple definitions of {!r}"
                                     .format(attr_name))

                self._delegations[attr_name] = attr_value
                # Remove class-level function added by ApplicationMeta.
                if hasattr(self.__class__, attr_name):
                    delattr(self.__class__, attr_name)

    def _init_gui(self):
        """Initialize the user interface."""
        vbox = gaupol.util.new_vbox(spacing=0)
        self._init_x_clipboard()
        self._init_window()
        self._init_actions()
        self._init_main_toolbar(vbox)
        self._init_paned(vbox)
        self._init_player_box(self.paned)
        self._init_notebook(self.paned)
        self.window.add(vbox)
        vbox.show_all()
        self._init_visibilities()

    def _init_main_toolbar(self, vbox):
        """Initialize the main toolbar."""
        self.main_toolbar = Gtk.Toolbar()
        style = self.main_toolbar.get_style_context()
        style.add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)
        toolbar_style = gaupol.conf.application_window.toolbar_style
        self.main_toolbar.set_style(toolbar_style.value)
        if sys.platform == "win32":
            self.main_toolbar.set_icon_size(Gtk.IconSize.MENU)
        gaupol.conf.connect_notify("application_window", "toolbar_style", self)
        gaupol.util.pack_start(vbox, self.main_toolbar)
        # win.open-main-files
        button = Gtk.MenuToolButton(
            label=_("Open"), icon_name="document-open")
        button.set_menu(Gtk.Menu())
        button.set_is_important(True)
        button.set_action_name("win.open-main-files")
        button.set_tooltip_text(_("Open main files"))
        callback = self._on_open_button_show_menu
        button.connect("show-menu", callback)
        self.main_toolbar.insert(button, -1)
        self.open_button = button
        # win.save-main-document
        button = Gtk.ToolButton(
            label=_("Save"), icon_name="document-save")
        button.set_is_important(True)
        button.set_action_name("win.save-main-document")
        button.set_tooltip_text(_("Save the current main document"))
        self.main_toolbar.insert(button, -1)
        self.main_toolbar.insert(Gtk.SeparatorToolItem(), -1)
        # win.undo-action
        button = Gtk.MenuToolButton(
            label=_("Undo"), icon_name="edit-undo")
        button.set_menu(Gtk.Menu())
        button.set_is_important(True)
        button.set_action_name("win.undo-action")
        button.set_tooltip_text(_("Undo the last action"))
        callback = self._on_undo_button_show_menu
        button.connect("show-menu", callback)
        self.main_toolbar.insert(button, -1)
        self.undo_button = button
        # win.redo-action
        button = Gtk.MenuToolButton(
            label=_("Redo"), icon_name="edit-redo")
        button.set_menu(Gtk.Menu())
        button.set_action_name("win.redo-action")
        button.set_tooltip_text(_("Redo the last undone action"))
        callback = self._on_redo_button_show_menu
        button.connect("show-menu", callback)
        self.main_toolbar.insert(button, -1)
        self.redo_button = button
        self.main_toolbar.insert(Gtk.SeparatorToolItem(), -1)
        # win.insert-subtitles
        button = Gtk.ToolButton(
            label=_("Insert"), icon_name="list-add")
        button.set_action_name("win.insert-subtitles")
        button.set_tooltip_text(_("Insert new subtitles"))
        self.main_toolbar.insert(button, -1)
        # win.remove-subtitles
        button = Gtk.ToolButton(
            label=_("Remove"), icon_name="list-remove")
        button.set_action_name("win.remove-subtitles")
        button.set_tooltip_text(_("Remove the selected subtitles"))
        self.main_toolbar.insert(button, -1)
        self.main_toolbar.insert(Gtk.SeparatorToolItem(), -1)
        # win.find-and-replace
        button = Gtk.ToolButton(
            label=_("Find"), icon_name="edit-find")
        button.set_action_name("win.find-and-replace")
        button.set_tooltip_text(_("Search for and replace text"))
        self.main_toolbar.insert(button, -1)
        # win.toggle-player
        button = Gtk.ToggleToolButton(
            label=_("Video"), icon_name="video-x-generic")
        button.set_action_name("win.toggle-player")
        button.set_tooltip_text(_("Show or hide the video player"))
        self.main_toolbar.insert(button, -1)
        # win.preview
        button = Gtk.ToolButton(
            label=_("Preview"), icon_name="media-playback-start")
        button.set_action_name("win.preview")
        button.set_tooltip_text(_("Preview from selected position with a video player"))
        self.main_toolbar.insert(button, -1)

    def _init_notebook(self, paned):
        """Initialize the notebook."""
        self.notebook = Gtk.Notebook()
        self.notebook.set_scrollable(True)
        self.notebook.set_show_border(False)
        self.notebook.set_show_tabs(False)
        self.notebook.drag_dest_set(flags=Gtk.DestDefaults.ALL,
                                    targets=None,
                                    actions=Gdk.DragAction.COPY)

        self.notebook.drag_dest_add_uri_targets()
        aeidon.util.connect(self, "notebook", "drag-data-received")
        aeidon.util.connect(self, "notebook", "page-reordered")
        callback = self._on_notebook_switch_page
        self.notebook.connect_after("switch-page", callback)
        orientation = Gtk.Orientation.HORIZONTAL
        self.notebook_separator = Gtk.Separator(orientation=orientation)
        vbox = gaupol.util.new_vbox(spacing=0)
        gaupol.util.pack_start(vbox, self.notebook_separator)
        overlay = Gtk.Overlay()
        overlay.add(self.notebook)
        self.statuslabel = gaupol.FloatingLabel()
        overlay.add_overlay(self.statuslabel)
        gaupol.util.pack_start_expand(vbox, overlay)
        paned.add2(vbox)

    def _init_paned(self, vbox):
        """Intialize the paned layout."""
        orientation = gaupol.conf.application_window.layout
        self.paned = Gtk.Paned(orientation=orientation)
        gaupol.util.pack_start_expand(vbox, self.paned)

    def _init_player_box(self, paned):
        """Initialize the video player horizontal box."""
        # This will actually be added to paned once a video is loaded.
        layout = gaupol.conf.application_window.layout
        if layout == Gtk.Orientation.HORIZONTAL:
            self.player_box = gaupol.util.new_vbox(spacing=0)
        if layout == Gtk.Orientation.VERTICAL:
            self.player_box = gaupol.util.new_hbox(spacing=0)

    def _init_visibilities(self):
        """Initialize visibilities of hideable widgets."""
        conf = gaupol.conf.application_window
        self.main_toolbar.set_visible(conf.show_main_toolbar)
        self.notebook_separator.set_visible(conf.show_main_toolbar)
        self.show_message(None)

    def _init_window(self):
        """Initialize the main window."""
        self.window = Gtk.ApplicationWindow(
            application=getattr(gaupol, "appman", None))
        self.window.set_show_menubar(True)
        self.window.set_icon_name("io.otsaloma.gaupol")
        Gtk.Window.set_default_icon_name("io.otsaloma.gaupol")
        self.window.set_default_size(*gaupol.conf.application_window.size)
        self.window.move(*gaupol.conf.application_window.position)
        if gaupol.conf.application_window.maximized:
            self.window.maximize()
        aeidon.util.connect(self, "window", "delete-event")
        aeidon.util.connect(self, "window", "window-state-event")
        gaupol.style.load_css(self.window)

    def _init_x_clipboard(self):
        """Initialize the desktop-wide, persistent clipboard."""
        atom = Gdk.atom_intern("CLIPBOARD", True)
        self.x_clipboard = Gtk.Clipboard.get(atom)
        self.x_clipboard.set_can_store(None)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Updating the application GUI."""

import aeidon
import gaupol

from gi.repository import Gdk


class UpdateAgent(aeidon.Delegate):

    """Updating the application GUI."""

    def _disable_widgets(self):
        """Make widgets insensitive and blank."""
        self.window.set_title("Gaupol")
        self.show_message(None)

    @aeidon.deco.export
    def flash_message(self, message, duration=6):
        """Show `message` in statuslabel for `duration` seconds."""
        self.statuslabel.flash_text(message, duration=duration)
        self.statuslabel.register_hide_event(self.window, "button-press-event")
        self.statuslabel.register_hide_event(self.window, "key-press-event")
        self.statuslabel.register_hide_event(self.window, "scroll-event")
        with aeidon.util.silent(AttributeError):
            self.statuslabel.register_hide_event(
                self.get_current_page().view, "button-press-event")

    @aeidon.deco.export
    def _on_activate_next_project_activate(self, *args):
        """Activate the project in the next tab."""
        self.notebook.next_page()

    @aeidon.deco.export
    def _on_activate_previous_project_activate(self, *args):
        """Activate the project in the previous tab."""
        self.notebook.prev_page()

    @aeidon.deco.export
    def _on_conf_application_window_notify_toolbar_style(self, *args):
        """Change the style of the main toolbar."""
        style = gaupol.conf.application_window.toolbar_style
        self.main_toolbar.set_style(style.value)

    @aeidon.deco.export
    def _on_move_tab_left_activate(self, *args):
        """Move the current tab to the left."""
        page = self.get_current_page()
        scroller = page.view.get_parent()
        index = self.pages.index(page)
        self.notebook.reorder_child(scroller, index-1)

    @aeidon.deco.export
    def _on_move_tab_right_activate(self, *args):
        """Move the current tab to the right."""
        page = self.get_current_page()
        scroller = page.view.get_parent()
        index = self.pages.index(page)
        self.notebook.reorder_child(scroller, index + 1)

    @aeidon.deco.export
    def _on_notebook_page_reordered(self, notebook, scroller, index):
        """Update the list of pages to match the new order."""
        view = scroller.get_child()
        page = [x for x in self.pages if x.view is view][0]
        self.pages.remove(page)
        self.pages.insert(index, page)
        self.update_gui()
        self.emit("pages-reordered", page, index)

    @aeidon.deco.export
    def _on_notebook_switch_page(self, notebook, pointer, index):
        """Update GUI for the page switched to."""
        if not self.pages: return
        self.update_gui()
        page = self.pages[index]
        page.view.grab_focus()
        self.emit("page-switched", page)

    @aeidon.deco.export
    def _on_view_move_cursor(self, *args):
        """Update GUI after moving cursor in the view."""
        self.update_gui()

    @aeidon.deco.export
    def _on_view_selection_changed(self, *args):
        """Update GUI after changing selection in the view."""
        self.update_gui()

    @aeidon.deco.export
    def _on_window_window_state_event(self, window, event):
        """Save window maximization."""
        state = event.new_window_state
        maximized = bool(state & Gdk.WindowState.MAXIMIZED)
        gaupol.conf.application_window.maximized = maximized

    @aeidon.deco.export
    def show_message(self, message):
        """Show `message` in the statuslabel or hide label with `None`."""
        self.statuslabel.set_text(message)

    def _update_actions(self, page):
        """Update sensitivities of all actions for page."""
        rows = []
        with aeidon.util.silent(AttributeError):
            rows = page.view.get_selected_rows()
        for name in self.window.list_actions():
            action = self.get_action(name)
            action.update_enabled(self, page, rows)

    @aeidon.deco.export
    def update_gui(self):
        """Update widget sensitivities and states for the current page."""
        page = self.get_current_page()
        self._update_actions(page)
        self._update_widgets(page)
        self.extension_manager.update_extensions(page)

    def _update_widgets(self, page):
        """Update states of all widgets for `page`."""
        if page is None:
            return self._disable_widgets()
        self.window.set_title(page.tab_label.get_text())
        self.notebook.set_menu_label_text(
            page.view.get_parent(), page.tab_label.get_text())
        tabs = len(self.pages) > 1 and not self.player_box.get_visible()
        self.notebook.set_show_tabs(tabs)
        self.get_action("set-edit-mode").set_state(str(page.edit_mode))
        self.get_action("set-framerate").set_state(str(page.project.framerate))
        if self.player is not None:
            layout = gaupol.conf.application_window.layout
            self.get_action("set-layout").set_state(str(layout))
        for field in gaupol.fields:
            col = getattr(page.view.columns, field.name)
            visible = page.view.get_column(col).get_visible()
            self.get_column_action(field).set_state(visible)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for warning when closing multiple documents."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("MultiCloseDialog",)


class MultiCloseDialog(Gtk.MessageDialog):

    """Dialog for warning when closing multiple documents."""

    def __init__(self, parent, application, pages):
        """Initialize a :class:`MultiCloseDialog` instance."""
        GObject.GObject.__init__(self,
                                 message_type=Gtk.MessageType.ERROR,
                                 text=_("Save changes to documents before closing?"),
                                 secondary_text=_("If you don't save, changes will be permanently lost."))

        self.application = application
        self._main_tree_view = Gtk.TreeView()
        self._main_vbox = gaupol.util.new_vbox(6)
        self.pages = tuple(pages)
        self._tran_tree_view = Gtk.TreeView()
        self._tran_vbox = gaupol.util.new_vbox(6)
        self._init_dialog(parent)
        self._init_main_tree_view()
        self._init_tran_tree_view()

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("Close _Without Saving"), Gtk.ResponseType.NO)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Save"), Gtk.ResponseType.YES)
        self.set_default_response(Gtk.ResponseType.YES)
        self.set_transient_for(parent)
        self.set_modal(True)
        aeidon.util.connect(self, self, "response")

    def _init_main_tree_view(self):
        """Initialize the main tree view."""
        store = self._init_tree_view(self._main_tree_view)
        for page in (x for x in self.pages if x.project.main_changed):
            store.append((page, True, page.get_main_basename()))
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scroller.add(self._main_tree_view)
        label = Gtk.Label(label=_("Select the _main documents you want to save:"))
        label.props.xalign = 0
        label.set_use_underline(True)
        label.set_mnemonic_widget(self._main_tree_view)
        gaupol.util.pack_start(self._main_vbox, label)
        gaupol.util.pack_start_expand(self._main_vbox, scroller)
        gaupol.util.pack_start_expand(self.get_message_area(), self._main_vbox)
        self._main_vbox.set_visible(len(store) > 0)
        if len(store) > 0:
            self._main_vbox.show_all()
            gaupol.util.scale_to_content(self._main_tree_view,
                                         min_nchar=30,
                                         max_nchar=60,
                                         min_nlines=2,
                                         max_nlines=6)

    def _init_tran_tree_view(self):
        """Initialize the translation tree view."""
        store = self._init_tree_view(self._tran_tree_view)
        for page in (x for x in self.pages if x.project.tran_changed):
            store.append((page, True, page.get_translation_basename()))
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scroller.add(self._tran_tree_view)
        label = Gtk.Label(label=_("Select the _translation documents you want to save:"))
        label.props.xalign = 0
        label.set_use_underline(True)
        label.set_mnemonic_widget(self._tran_tree_view)
        gaupol.util.pack_start(self._tran_vbox, label)
        gaupol.util.pack_start_expand(self._tran_vbox, scroller)
        gaupol.util.pack_start_expand(self.get_message_area(), self._tran_vbox)
        self._tran_vbox.set_visible(len(store) > 0)
        if len(store) > 0:
            self._tran_vbox.show_all()
            gaupol.util.scale_to_content(self._tran_tree_view,
                                         min_nchar=30,
                                         max_nchar=60,
                                         min_nlines=2,
                                         max_nlines=6)

    def _init_tree_view(self, tree_view):
        """Initialize `tree_view` and return its model."""
        tree_view.set_headers_visible(False)
        tree_view.set_enable_search(False)
        store = Gtk.ListStore(object, bool, str)
        tree_view.set_model(store)
        selection = tree_view.get_selection()
        selection.set_mode(Gtk.SelectionMode.SINGLE)
        renderer = Gtk.CellRendererToggle()
        renderer.props.activatable = True
        renderer.connect("toggled", self._on_tree_view_cell_toggled, store)
        column = Gtk.TreeViewColumn("", renderer, active=1)
        tree_view.append_column(column)
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("", renderer, text=2)
        tree_view.append_column(column)
        return store

    def _on_response(self, dialog, response):
        """Save the selected documents and close pages."""
        if response == Gtk.ResponseType.YES:
            for page in self.pages:
                self._save_and_close_page(page)
        if response == Gtk.ResponseType.NO:
            for page in self.pages:
                self.application.close(page, confirm=False)

    def _on_tree_view_cell_toggled(self, renderer, path, store):
        """Toggle save document check button value."""
        store[path][1] = not store[path][1]
        store = self._main_tree_view.get_model()
        mains = [x for x in store if x[1]]
        store = self._tran_tree_view.get_model()
        trans = [x for x in store if x[1]]
        sensitive = bool(mains or trans)
        self.set_response_sensitive(Gtk.ResponseType.YES, sensitive)

    @aeidon.deco.silent(gaupol.Default)
    def _save_and_close_page(self, page):
        """Save the selected documents and close `page`."""
        store = self._main_tree_view.get_model()
        pages = [x for x in store if x[0] is page]
        if pages and pages[0][1]:
            self.application.save_main(page)
        store = self._tran_tree_view.get_model()
        pages = [x for x in store if x[0] is page]
        if pages and pages[0][1]:
            self.application.save_translation(page)
        self.application.close(page, confirm=False)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Names and codes for locales and conversions between them.

Locale codes are of form ``aa[_BB][@Cccc]``, where ``aa`` is a language code,
``BB`` a country code and ``Cccc`` a script. See :mod:`aeidon.languages`,
:mod:`aeidon.countries` and :mod:`aeidon.scripts` for details.
"""

import aeidon
import os

from aeidon.i18n import _


def code_to_country(code):
    """Convert locale `code` to localized country name or ``None``."""
    if len(code) < 5: return None
    return aeidon.countries.code_to_name(code[-2:])

def code_to_language(code):
    """Convert locale `code` to localized language name."""
    if len(code) < 2: return None
    return aeidon.languages.code_to_name(code[:2])

def code_to_name(code):
    """Convert locale `code` to localized name."""
    if len(code) < 5:
        return code_to_language(code)
    return (_("{language} ({country})")
            .format(language=code_to_language(code),
                    country=code_to_country(code)))

@aeidon.deco.once
def get_system_code():
    """Return the locale code preferred by system or ``None``."""
    import locale
    return locale.getdefaultlocale()[0]

@aeidon.deco.once
def get_system_modifier():
    """Return the system default script modifier or ``None``."""
    for name in ("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG"):
        value = os.environ.get(name, None)
        if value and value.count("@") == 1:
            i = value.index("@")
            return value[i+1:i+5]
    # No script modifier found implies the language default script.
    return None



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Widget to display subtitle data in the form of a list."""

import aeidon
import gaupol
import re

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("View",)


class View(Gtk.TreeView):

    """
    Widget to display subtitle data in the form of a list.

    :ivar columns: Enumeration for columns currently shown

       The values of the enumeration items correspond to the column indices and
       are updated when columns are added, removed or reordered. Note that
       these indices are not necessarily the same as the column indices in the
       underlying :class:`Gtk.ListStore` data model.
    """

    def __init__(self, edit_mode):
        """Initialize a :class:`View` instance."""
        GObject.GObject.__init__(self)
        self._active_col_name = ""
        self._calc = aeidon.Calculator()
        self.columns = aeidon.Enumeration()
        self._selection_changed_handlers = {}
        self._init_signal_handlers()
        self._init_props(edit_mode)

    def connect_selection_changed(self, callback):
        """
        Connect to the "changed" signal of selection.

        Using this instead of the tree selection's own ``connect`` method
        allows signal handlers to be blocked while in the process of selecting
        multiple rows, which should be significantly faster if actually doing
        something in the signal handler.

        Return signal handler ID.
        """
        selection = self.get_selection()
        handler_id = selection.connect("changed", callback)
        self._selection_changed_handlers[handler_id] = callback
        return handler_id

    def disconnect_selection_changed(self, callback):
        """Disconnect from the "changed" signal of selection."""
        selection = self.get_selection()
        for handler_id in list(self._selection_changed_handlers):
            if self._selection_changed_handlers[handler_id] is callback:
                selection.handler_disconnect(handler_id)
                self._selection_changed_handlers.pop(handler_id)

    def get_focus(self):
        """Return the row and column of the current focus."""
        path, col = self.get_cursor()
        row = gaupol.util.tree_path_to_row(path)
        if col is not None:
            col = self.get_columns().index(col)
        return row, col

    def get_header_label(self, field=None, title=None):
        """Return a column header label from `text`."""
        title = title or field.label
        label = Gtk.Label(label=title)
        # Avoid column resizing by "preallocating"
        # sufficient width for usual expected data.
        if field == gaupol.fields.NUMBER:
            size_label = Gtk.Label(label="8888")
            size_label.show()
            width = size_label.get_preferred_width()[1]
            label.set_size_request(width, -1)
        if field == gaupol.fields.DURATION:
            size_label = Gtk.Label(label="88.888")
            size_label.show()
            width = size_label.get_preferred_width()[1]
            label.set_size_request(width, -1)
        label.set_halign(Gtk.Align.START)
        label.show()
        return label

    def _get_renderer(self, field, edit_mode):
        """Initialize and return a new cell renderer for `field`."""
        font = gaupol.util.get_font()
        if field == gaupol.fields.NUMBER:
            renderer = gaupol.IntegerCellRenderer()
            renderer.props.editable = False
            renderer.props.xalign = 1
        elif field.is_position:
            if edit_mode == aeidon.modes.TIME:
                if field == gaupol.fields.DURATION:
                    renderer = gaupol.FloatCellRenderer()
                else: renderer = gaupol.TimeCellRenderer()
            elif edit_mode == aeidon.modes.FRAME:
                renderer = gaupol.IntegerCellRenderer()
            renderer.props.editable = True
            renderer.props.xalign = 1
        elif field.is_text:
            renderer = gaupol.MultilineCellRenderer()
            renderer.props.editable = True
            renderer.props.xalign = 0
        renderer.props.font = font
        renderer.props.yalign = 0
        renderer.props.xpad = 4
        renderer.props.ypad = 4
        return renderer

    def get_selected_rows(self):
        """Return a sequence of selected rows."""
        paths = self.get_selection().get_selected_rows()[1]
        return tuple(gaupol.util.tree_path_to_row(x) for x in paths)

    def _init_cell_data_functions(self):
        """Initialize functions to automatically update cell data."""
        # Set the data in the number column automatically.
        def set_number(column, renderer, store, itr, data):
            path = store.get_path(itr)
            row = gaupol.util.tree_path_to_row(path)
            renderer.props.text = str(row + 1)
        column = self.get_column(self.columns.NUMBER)
        renderer = column.get_cells()[0]
        column.set_cell_data_func(renderer, set_number, None)
        if gaupol.conf.editor.use_zebra_stripes:
            column = self.get_column(self.columns.MAIN_TEXT)
            renderer = column.get_cells()[0]
            callback = self._on_renderer_set_background
            column.set_cell_data_func(renderer, callback, None)

    def _init_columns(self, edit_mode):
        """Initialize the tree view columns."""
        visible_fields = gaupol.conf.editor.visible_fields
        for field in gaupol.conf.editor.field_order:
            renderer = self._get_renderer(field, edit_mode)
            column = Gtk.TreeViewColumn(field.label, renderer, text=field)
            column.gaupol_id = field.name.lower()
            self.append_column(column)
            column.set_clickable(True)
            column.set_resizable(True)
            column.set_reorderable(True)
            column.set_expand(field.is_text)
            column.set_visible(field in visible_fields)
            label = self.get_header_label(field)
            label.set_tooltip_text(field.tooltip)
            column.set_widget(label)

    def _init_props(self, edit_mode):
        """Initialize properties."""
        if edit_mode == aeidon.modes.TIME:
            columns = (int, str, str, float, str, str)
        if edit_mode == aeidon.modes.FRAME:
            columns = (int, int, int, int, str, str)
        store = Gtk.ListStore(*columns)
        self.set_model(store)
        self._init_columns(edit_mode)
        self._init_cell_data_functions()
        self.set_name("gaupol-view")
        self.set_headers_visible(True)
        self.set_rubber_banding(True)
        selection = self.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self._init_search()

    def _init_search(self):
        """Initialize the interactive search properties."""
        self.set_enable_search(True)
        self.set_search_column(self.columns.NUMBER)
        self.set_search_equal_func(self._search_equals, None)

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        aeidon.util.connect(self, self, "columns-changed")
        aeidon.util.connect(self, self, "cursor-changed")
        aeidon.util.connect(self, self, "key-press-event")
        gaupol.conf.connect_notify("editor", "custom_font", self)
        gaupol.conf.connect_notify("editor", "length_unit", self)
        gaupol.conf.connect_notify("editor", "show_lengths_cell", self)
        gaupol.conf.connect_notify("editor", "use_custom_font", self)

    def is_position_column(self, col):
        """Return ``True`` if `col` is a position column."""
        return col in (self.columns.START,
                       self.columns.END,
                       self.columns.DURATION)

    def is_text_column(self, col):
        """Return True if `col` is a text column."""
        return col in (self.columns.MAIN_TEXT,
                       self.columns.TRAN_TEXT)

    def _on_conf_editor_notify_custom_font(self, *args):
        """Apply the new font to all columns."""
        font = gaupol.util.get_font()
        for column in self.get_columns():
            renderer = column.get_cells()[0]
            if hasattr(renderer.props, "font"):
                renderer.props.font = font
        self.columns_autosize()

    def _on_conf_editor_notify_length_unit(self, *args):
        """Repaint the cells to update line length display."""
        if gaupol.conf.editor.show_lengths_cell:
            self.columns_autosize()

    def _on_conf_editor_notify_show_lengths_cell(self, *args):
        """Repaint the cells to update line length display."""
        self.columns_autosize()

    def _on_conf_editor_notify_use_custom_font(self, *args):
        """Apply the new font to all columns."""
        font = gaupol.util.get_font()
        for column in self.get_columns():
            renderer = column.get_cells()[0]
            if hasattr(renderer.props, "font"):
                renderer.props.font = font
        self.columns_autosize()

    def _on_columns_changed(self, *args):
        """Reset the columns enumeration to match new column layout."""
        count_changed = (len(self.columns) != len(self.get_columns()))
        self._reset_columns()
        field_order = []
        for column in self.get_columns():
            attribute = column.gaupol_id.upper()
            item = aeidon.EnumerationItem()
            setattr(self.columns, attribute, item)
            if hasattr(gaupol.fields, attribute):
                field = getattr(gaupol.fields, attribute)
                field_order.append(field)
        if not count_changed:
            gaupol.conf.editor.field_order = field_order

    def _on_cursor_changed(self, *args):
        """Update the column header labels to reflect changed focus."""
        self.update_headers()

    def _on_key_press_event(self, widget, event):
        """Handle various special-case key combinations."""
        # Disable Ctrl+PageUp/PageDown to allow them to be
        # used solely for navigation between notebook tabs.
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            if event.keyval in (Gdk.KEY_Page_Up, Gdk.KEY_Page_Down):
                return widget.stop_emission("key-press-event")
        # Use interactive search for a subtitle number or time
        # only if valid string keys have been pressed.
        self.set_enable_search(event.string.isdigit() or
                               event.string == ":" or
                               event.string == ".")

    def _on_renderer_set_background(self, column, renderer, store, itr, data):
        """Set zerba-striped backgrounds for all columns."""
        path = self.get_model().get_path(itr)
        row = gaupol.util.tree_path_to_row(path)
        color = (gaupol.util.get_zebra_color(self)
                 if row % 2 == 0 else None)

        for column in self.get_columns():
            for renderer in column.get_cells():
                renderer.props.cell_background_rgba = color

    def _reset_columns(self):
        """Recreate the columns enumeration and set all items to ``None``."""
        self.columns = aeidon.Enumeration()
        self.columns.NUMBER = None
        self.columns.START = None
        self.columns.END = None
        self.columns.DURATION = None
        self.columns.MAIN_TEXT = None
        self.columns.TRAN_TEXT = None

    def scroll_to_row(self, row):
        """Scroll view until `row` is visible."""
        self.scroll_to_cell(path=row,
                            column=None,
                            use_align=False,
                            row_align=0.5,
                            col_align=0)

    def _search_equals(self, store, column, key, itr, data):
        """Return ``False`` if `key` matches either subtitle number or time."""
        path = store.get_path(itr)
        row = gaupol.util.tree_path_to_row(path)
        if key.count(":") == 0:
            # Search for subtitle number.
            try:
                return row != (int(key) - 1)
            except ValueError:
                return False
        col = gaupol.conf.editor.field_order.index(gaupol.fields.START)
        hours = minutes = seconds = 0
        if key.count(":") == 1 and key.startswith(":"):
            # Search for time of form ':MM'
            match = re.search(r"^:(\d{1,2})$", key)
            if match is None: return False
            minutes = int(match.group(1))
        if key.count(":") == 1 and not key.startswith(":"):
            # Search for time of form 'MM:[SS.SSS]'
            match = re.search(r"^(\d{1,2}):([\d.]+)?$", key)
            if match is None: return False
            minutes = int(match.group(1))
            try:
                seconds = float(match.group(2) or "00.000")
            except ValueError:
                return False
        if key.count(":") == 2:
            # Search for time of form '[HH]:MM:[SS.SSS]'
            match = re.search(r"^(\d{,2}):(\d{1,2}):([\d.]+)?$", key)
            if match is None: return False
            hours = int(match.group(1) or "00")
            minutes = int(match.group(2))
            try:
                seconds = float(match.group(3) or "00.000")
            except ValueError:
                return False
        time_key = "{:02d}:{:02d}:{:06.3f}".format(hours, minutes, seconds)
        if not self._calc.is_valid_time(time_key): return False
        time_iter = store[row][col]
        if not ":" in time_iter: return False
        try:
            time_next = store[row+1][col]
        except IndexError:
            time_next = "99:59:59.999"
        return not (self._calc.time_to_seconds(time_iter) <=
                    self._calc.time_to_seconds(time_key) <
                    self._calc.time_to_seconds(time_next))

    def select_rows(self, rows):
        """Select `rows`, clearing previous selection."""
        # Avoid sending more than one 'changed' signal.
        selection = self.get_selection()
        for handler_id in self._selection_changed_handlers:
            selection.handler_block(handler_id)
        selection.unselect_all()
        for lst in aeidon.util.get_ranges(rows):
            start = gaupol.util.tree_row_to_path(lst[0])
            end = gaupol.util.tree_row_to_path(lst[-1])
            selection.select_range(start, end)
        for handler_id in self._selection_changed_handlers:
            selection.handler_unblock(handler_id)
        selection.emit("changed")

    def set_focus(self, row, col=None):
        """Set the focus to `row` (-1 for last), `col`."""
        if row == -1:
            row = len(self.get_model()) - 1
        if col is not None:
            col = self.get_column(col)
        path = gaupol.util.tree_row_to_path(row)
        self.set_cursor(path, col, start_editing=False)
        if col is not None:
            self.update_headers()

    def update_headers(self):
        """Update the attributes of all column header labels."""
        fcol = self.get_focus()[1]
        acol = getattr(self.columns, self._active_col_name, None)
        if fcol == acol: return
        self._active_col_name = ""
        if acol is not None:
            label = self.get_column(acol).get_widget()
            text = GLib.markup_escape_text(label.get_text())
            label.set_markup("{}".format(text))
        if fcol is not None:
            label = self.get_column(fcol).get_widget()
            text = GLib.markup_escape_text(label.get_text())
            label.set_markup("<i>{}</i>".format(text))
            self._active_col_name = self.columns[fcol].name



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Text markup for the SubViewer 2.0 format."""

import aeidon

__all__ = ("SubViewer2",)


class SubViewer2(aeidon.markups.SubRip):

    """
    Text markup for the SubViewer 2.0 format.

    SubViewer 2.0 format is assumed to contain the same markup as SubRip.
    """

    format = aeidon.formats.SUBVIEWER2



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Creating and removing temporary files and directories."""

import aeidon
import atexit
import os
import tempfile

_paths = []


def create(suffix=""):
    """Create a new temporary file and return its path."""
    handle, path = tempfile.mkstemp(suffix, "gaupol-")
    # Avoid PermissionErrors etc. on Windows by closing
    # the handle and returning only the path which will
    # be opened and closed separately.
    os.close(handle)
    _paths.append(path)
    return path

def create_directory(suffix=""):
    """Create a new temporary directory and return its path."""
    path = tempfile.mkdtemp(suffix, "gaupol-")
    _paths.append(path)
    return path

def remove(path):
    """Remove temporary file or directory at `path`."""
    if os.path.isfile(path):
        with aeidon.util.silent(OSError):
            os.remove(path)
    if os.path.isdir(path):
        for name in os.listdir(path):
            remove(os.path.join(path, name))
        with aeidon.util.silent(OSError):
            os.rmdir(path)

def remove_all():
    """Remove all temporary files and directories."""
    for path in _paths:
        remove(path)

atexit.register(remove_all)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Help actions for :class:`gaupol.Application`."""

import gaupol

class BrowseDocumentationAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "browse-documentation")
        self.action_group = "safe"

class ReportABugAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "report-a-bug")
        self.action_group = "safe"

class ViewAboutDialogAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "view-about-dialog")
        self.action_group = "safe"

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Closing pages and quitting Gaupol."""

import aeidon
import gaupol
import os

from aeidon.i18n   import _
from gi.repository import Gtk


class CloseAgent(aeidon.Delegate):

    """Closing pages and quitting Gaupol."""

    @aeidon.deco.export
    def close(self, page, confirm=True):
        """Close `page` after asking to save its documents."""
        if confirm:
            self._confirm_close(page)
        if not page in self.pages: return
        index = self.pages.index(page)
        if self.notebook.get_current_page() == index:
            self.notebook.next_page()
        self.notebook.remove_page(index)
        self.pages.remove(page)
        self.update_gui()
        self.emit("page-closed", page)

    @aeidon.deco.export
    def close_all(self, confirm=True):
        """Close all pages after asking to save their documents."""
        if confirm and sum(len(self._need_confirmation(
                x)) for x in self.pages) > 1:
            return self._confirm_close_multiple(tuple(self.pages))
        while self.pages:
            self.close(self.pages[-1], confirm=confirm)

    def _confirm_close(self, page):
        """Close `page` after asking to save its documents."""
        docs = self._need_confirmation(page)
        if len(docs) == 0: return
        if len(docs) == 1:
            self._confirm_close_document(page, docs[0])
        if len(docs) > 1:
            self._confirm_close_multiple((page,))

    def _confirm_close_document(self, page, doc):
        """Close `page` after asking to save `doc`."""
        title = _('Save changes to document "{}" before closing?')
        title = title.format(page.get_basename(doc))
        message = _("If you don't save, changes will be permanently lost.")
        dialog = gaupol.WarningDialog(self.window, title, message)
        dialog.add_button(_("Close _Without Saving"), Gtk.ResponseType.NO)
        dialog.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        dialog.add_button(_("_Save"), Gtk.ResponseType.YES)
        dialog.set_default_response(Gtk.ResponseType.YES)
        response = gaupol.util.flash_dialog(dialog)
        gaupol.util.raise_default(not response in (
            Gtk.ResponseType.YES, Gtk.ResponseType.NO))
        if response == Gtk.ResponseType.YES:
            self.save(page, doc)

    def _confirm_close_multiple(self, pages):
        """Close `pages` after asking to save their documents."""
        dialog = gaupol.MultiCloseDialog(self.window, self, pages)
        response = gaupol.util.flash_dialog(dialog)
        gaupol.util.raise_default(not response in (
            Gtk.ResponseType.YES, Gtk.ResponseType.NO))

    def _need_confirmation(self, page):
        """Return documents in `page` with unsaved changes."""
        docs = []
        if page.project.main_changed:
            docs.append(aeidon.documents.MAIN)
        elif page.project.main_file is not None:
            if not os.path.isfile(page.project.main_file.path):
                docs.append(aeidon.documents.MAIN)
        if page.project.tran_changed:
            docs.append(aeidon.documents.TRAN)
        elif page.project.tran_file is not None:
            if not os.path.isfile(page.project.tran_file.path):
                docs.append(aeidon.documents.TRAN)
        return tuple(docs)

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_close_all_projects_activate(self, *args):
        """Close all open projects."""
        self.close_all()

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_close_project_activate(self, *args):
        """Close project."""
        self.close(self.get_current_page())

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_page_close_request(self, page, *args):
        """Close project."""
        self.close(page)

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_quit_activate(self, *args):
        """Quit Gaupol."""
        self.quit()

    @aeidon.deco.export
    def _on_window_delete_event(self, *args):
        """Quit Gaupol."""
        try:
            self.quit()
        except gaupol.Default:
            pass
        # Let appman destroy the window if confirmation given
        # and once any events being processed have completed.
        # https://github.com/otsaloma/gaupol/issues/54
        return True

    @aeidon.deco.export
    def quit(self, confirm=True):
        """Quit Gaupol."""
        self.emit("quit")
        if confirm and sum(len(self._need_confirmation(
                x)) for x in self.pages) > 1:
            self._confirm_close_multiple(tuple(self.pages))
        elif confirm:
            for page in self.pages:
                if self._need_confirmation(page):
                    self._confirm_close(page)
        self.extension_manager.teardown_extensions()
        if not gaupol.conf.application_window.maximized:
            conf = gaupol.conf.application_window
            conf.size = list(self.window.get_size())
            conf.position = list(self.window.get_position())
        if hasattr(gaupol, "appman"):
            gaupol.appman.remove_window(self.window)
